
// ====================================
// Scanner's AspectJ - Utilities aspect
// ====================================

package redhorizon.exception;

import redhorizon.strings.ProgramStrings;
import redhorizon.utilities.DataSource;
import redhorizon.utilities.DataSink;
import redhorizon.utilities.FilePointer;
import redhorizon.utilities.FileUtility;

import java.awt.image.BufferedImage;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.text.ParseException;
import java.util.concurrent.BlockingQueue;
import javax.imageio.ImageIO;

/**
 * Currently, this aspect's only task is to soften the several checked
 * exceptions that occur within the classes of this package.
 * 
 * @author Emanuel Rabina
 */
public aspect CoreExceptionDelegator {

	/**
	 * Soften <code>FileNotFoundException</code> and <code>IOException</code>
	 * types on the methods of file-reading/writing classes such as
	 * {@link FileUtility} and {@link FilePointer}. 
	 */
	pointcut fileUtilMethods():
		execution(public static * FileUtility.*(..));

	pointcut filePointMethods():
		execution(public FilePointer.new(..)) ||
		execution(public * FilePointer.*(..));

	pointcut fileImageRead():
		call(public static BufferedImage ImageIO.read(..));

	pointcut fileImageWrite():
		call(public static boolean ImageIO.write(..));

	declare soft: FileNotFoundException:
		fileUtilMethods() || filePointMethods();
	declare soft: IOException:
		fileUtilMethods() || filePointMethods() || fileImageRead() || fileImageWrite();

	/**
	 * Soften the many exceptions that can occur during the operations on a
	 * {@link DataStream}, {@link DataSink}, and {@link DataSource}.
	 */
	pointcut streamRead():
		execution(public * DataSource.*(..));

	pointcut streamWrite():
		execution(public * DataSink.*(..));

	declare soft: IOException: streamRead() || streamWrite();

	/**
	 * Soften interruption exceptions on the blocking queue methods.
	 */
	pointcut blockQueue():
		call(public void BlockingQueue.put(..)) || call (public * BlockingQueue.take());

	declare soft: InterruptedException: blockQueue();

	/**
	 * Soften the parser exception in the program strings / locale manager.
	 */
	pointcut stringParse():
		execution(public static float ProgramStrings.percentToFloat(String));

	declare soft: ParseException: stringParse();
}
