
// ==================================================================
// Scanner's Java - Type identifier for implementations of game files
// ==================================================================

package redhorizon.filetypes;

import redhorizon.strings.Common_Filetypes;
import redhorizon.strings.ProgramStrings;
import redhorizon.utilities.FilePointer;

import java.io.File;

/**
 * Abstract class and type identifier for actual game content files.  These
 * types of files act as implementations of the matching file type, including
 * decoding/decompression routines (if necessary) to transform the file from the
 * binary data format to a form useable by the game.
 * 
 * @author Emanuel Rabina
 */
public abstract class FileImplementation {

	protected final String name;
	protected FilePointer file;
	private boolean cleared;

	/**
	 * Constructor, assigns this representation a name.
	 * 
	 * @param name Name of the file being represented by the class.
	 * @param file <code>FilePointer</code> to the matching file on disk.
	 */
	protected FileImplementation(String name, FilePointer file) {

		this.name = name.contains(File.separator) ?
				name.substring(name.lastIndexOf(File.separator) + 1) : name;
		this.file = file;

		// Close file links on JVM shutdown
		Runtime.getRuntime().addShutdownHook(new Thread() {
			public void run() {
				if (!cleared) {
					FileImplementation.this.file.close();
				}
			}
		});
	}

	/**
	 * Returns whether the filenames of 2 <code>FileImplementation</code>s
	 * match.  Unique filenames are a requirement of Red Horizon.
	 * 
	 * @param other The other file to compare against.
	 * @return <code>true</code> if the filenames match, <code>false</code>
	 * 		   otherwise.
	 */
	public boolean equals(Object other) {

		return (other instanceof FileImplementation) && name.equals(((FileImplementation)other).name);
	}

	/**
	 * Closes the file link during a garbage collection cycle, instead of at JVM
	 * shutdown.
	 */
	protected void finalize() {

		file.close();
		file = null;
		cleared = true;
	}

	/**
	 * Returns the name of the file.
	 * 
	 * @return The filename this file was built from.
	 */
	public String getFilename() {

		return name;
	}

	/**
	 * Hashcode algorithm for files.  It is simply the hashcode for the file's
	 * name, meaning that filenames must be unique across the game.  For the
	 * requirements of Red Horizon, this is sufficient.
	 * 
	 * @return The hashcode for the file's name string.
	 */
	public int hashCode() {

		return name.hashCode();
	}

	/**
	 * Loads the data of this file from the current <code>FilePointer</code>.
	 * Body processing is only done if it is required by the file format.
	 * Otherwise, body information isn't read and decoded/decompressed until the
	 * implementation is used/played (eg: {@link Streaming} files).
	 */
	public abstract void readFromFile() throws UnsupportedFileTypeException;

	/**
	 * Saves the current data stored in this implementation, to the current
	 * <code>FilePointer</code> (set to the location of the save data).  Note
	 * that not all file types support saving, and so calling this method on
	 * such a type will throw a <code>UnsupportedOperationException</code>.
	 * 
	 * @throws UnsupportedOperationException For if the implementation does not
	 * 		   support saving to a file.
	 */
	public void saveToFile() throws UnsupportedOperationException {

		throw new UnsupportedOperationException(ProgramStrings.getText(
				Common_Filetypes.FILE_NO_SAVE, getClass().getName()));
	}
}
