
// =========================================
// Scanner's Java - Palette format interface
// =========================================

package redhorizon.filetypes;

/**
 * Interface for palette type file formats (eg: PAL, err... anything else?).<br>
 * <br>
 * At the moment, I think it best to just have these formats return either the
 * RGB or RGBA values instead of the palette data itself.  Unless I find a way
 * to re-create paletted textures in OpenGL (a feature which has since been
 * deemed obsolete by modern-day graphics cards), colour data will only be
 * returned.
 * 
 * @author Emanuel Rabina
 */
public interface PaletteFile {

	/**
	 * Applies an alpha scheme to the palette, making further requests from
	 * {@link #getRGBA(int)} return transparency values at specified indices.
	 * The alpha can either be a complete replacement for a colour (all 4
	 * replacement values are between 0-255) or just certain colours (those
	 * which aren't intended to override are set to -1).  Checcks for this are
	 * performed by the implementation, and handled appropriately.
	 * 
	 * @param palalpha <code>PaletteAlpha</code> containing an alpha scheme to
	 * 				   apply.
	 */
	public void applyAlpha(PaletteAlpha palalpha);

	/**
	 * Performs a palette shift so that future requests to the palette return
	 * the altered values instead.  At the moment, only side-specific changes
	 * are supported; shifts for other colours such as the water blue and
	 * glowing red have not yet been implemented.
	 * 
	 * @param palshift <code>PaletteShift</code> for a new palette colour
	 * 				   scheme.
	 */
	public void applyShift(PaletteShift palshift);

	/**
	 * Returns a modified filename, which is a concatenation of the palette
	 * filename, with whatever modifications have been applied to the file.
	 * 
	 * @return <code>(filename)[_(shift used)_(alpha_used)]</code>
	 */
	public String getFilename();

	/**
	 * Returns the RGB components of the palette entry at the given index.  In
	 * the returned array, the first value is R, the second G, and the third B.
	 * No alpha values are specified.  To obtain alpha values, use
	 * {@link #getRGBA(int)}.
	 * 
	 * @param index Entry number to look up in the palette.
	 * @return 3-element <code>byte</code>[] containing the RGB values.
	 */
	public byte[] getRGB(int index);

	/**
	 * Same as {@link #getRGB(int)} except that the returned array has a 4th
	 * <code>byte</code> value for the alpha component of the palette entry.
	 * 
	 * @param index Entry number to look up in the palette.
	 * @return 4-element <code>byte</code>[] containing the RGBA values.
	 */
	public byte[] getRGBA(int index);
}
