
// ================================================
// Scanner's Java - Interface for streaming formats
// ================================================

package redhorizon.filetypes;

import redhorizon.utilities.DataStream;

/**
 * Classes implementing this format should return <code>null</code> at the
 * methods used to normally get the data (eg: <code>getFrames()</code> or
 * <code>getSound()</code>) and instead use this class' {@link #getPipe()}
 * method to return the {@link DataStream}, to retrieve audio/video data in a
 * streaming manner.<br>
 * <br>
 * Data that is written to or read from the {@link DataStream} obtained from a
 * streaming file should be ordered by the priorities described by the constants
 * in this interface (lower number = should be closer to first).  This should
 * alleviate the confusion as to what sort of data comes first (if there are
 * multiple types) in the stream.  Note that these are not actual positions in
 * the stream where these parts are; just priority guides.<br>
 * <br>
 * Implementations should still write the methods required to decode and stream
 * the data, but this should then be wrapped-up in a <code>Runnable</code> (to
 * become a <code>Thread</code> later) and passed to the {@link DataStream}
 * class via it's {@link DataStream#setAlgorithm(Runnable)} method.
 * 
 * @author Emanuel Rabina
 */
public interface Streaming {

	public static final int PIPE_IMAGEDATA = 1;
	public static final int PIPE_SOUNDDATA = 2;

	/**
	 * Returns the blockable pipe used to implement streaming between this
	 * implementation's decoding thread and the media type's (video?
	 * long sound?) display/playback methods.<br>
	 * <br>
	 * Writes should be performed as fast as possible (up to the limiting size
	 * of the <code>DataStream</code>), whereas read speed should be deduced
	 * from some method in the media (eg: videos, use the framerate).
	 * 
	 * @return A <code>DataStream</code> in which the data will be streamed.
	 */
	public DataStream getPipe();
}
