
// ========================================
// Scanner's Java - AUD file representation
// ========================================

package redhorizon.filetypes.aud;

import redhorizon.filetypes.FileImplementation;
import redhorizon.filetypes.SoundFile;
import redhorizon.utilities.BufferUtility;
import redhorizon.utilities.CNCCodec;
import redhorizon.utilities.FilePointer;

import java.nio.ByteBuffer;

/**
 * Implementation of the AUD files used in Red Alert and Tiberium Dawn.  An AUD
 * file is the sound format of choice for these games, compressed using one of 2
 * schemes: IMA-ADPCM and WS-ADPCM.  The latter being a Westwood proprietary
 * format.
 * 
 * @author Emanuel Rabina
 */
public class AudFile extends FileImplementation implements SoundFile {

	static final byte TYPE_IMA_ADPCM = 99;
	static final byte TYPE_WS_ADPCM  = 1;
	static final byte FLAG_16BIT  = 0x02;
	static final byte FLAG_STEREO = 0x01;

	AudHeader audheader;
	private ByteBuffer audsound;

	/**
	 * Constructor, assigns the filename to be the name of the file, and
	 * generates a new <code>FilePointer</code> using that filename.
	 * 
	 * @param name The file to be parsed.
	 */
	public AudFile(String name) {

		this(name, new FilePointer(name));
	}

	/**
	 * Constructor, takes a name for the name of the file and the
	 * <code>FilePointer</code> as the location of the file.  Body processing
	 * only occurs on larger AUD files, such as music tracks.
	 * 
	 * @param name The name of this file.
	 * @param file Pointer to the file to parse.
	 */
	public AudFile(String name, FilePointer file) {

		super(name, file);
	}

	/**
	 * @inheritDoc
	 */
	public int bitrate() {

		return ((audheader.flags & FLAG_16BIT) != 0) ? 16 : 8;
	}

	/**
	 * @inheritDoc
	 */
	public int channels() {

		return ((audheader.flags & FLAG_STEREO) != 0) ? 2 : 1;
	}

	/**
	 * @inheritDoc
	 */
	public int frequency() {

		return audheader.frequency & 0xffff;
	}

	/**
	 * @inheritDoc
	 */
	public ByteBuffer getSound() {

		return audsound;
	}

	/**
	 * @inheritDoc
	 */
	public void readFromFile() {

		// Init header
		audheader = new AudHeader(file);

		// Decompresses the compressed AUD file data
		audsound = BufferUtility.newByteBuffer(audheader.datasize);
		byte pcmtype = audheader.type;
		int[] update = { 0, 0 };

		// Decode all remaining chunks
		while (file.remaining() > 0) {

			// Sound chunk
			AudChunk audchunk = new AudChunk(file);
			ByteBuffer sourcebytes = file.readBytes(audchunk.filesize);
			ByteBuffer destbytes   = BufferUtility.newByteBuffer(audchunk.datasize);

			// Select the proper decompression scheme and decompress
			if (pcmtype == TYPE_WS_ADPCM) {
				CNCCodec.decompress8bitWSADPCM(sourcebytes, destbytes);
			}
			else if (pcmtype == TYPE_IMA_ADPCM) {
				CNCCodec.decompress16bitIMAADPCM(sourcebytes, destbytes, update);
			}

			// Append chunk to complete sound
			audsound.put(destbytes);
		}
		audsound.rewind();
	}
}
