
// ===================================
// Scanner's Java - Streaming AUD file
// ===================================

package redhorizon.filetypes.aud;

import redhorizon.filetypes.Streaming;
import redhorizon.utilities.BufferUtility;
import redhorizon.utilities.CNCCodec;
import redhorizon.utilities.DataStream;
import redhorizon.utilities.FilePointer;

import java.nio.ByteBuffer;

/**
 * Like the {@link AudFile}, except that this is used for larger AUD files and
 * implements streaming, to reduce the memory usage of the audio data.
 * 
 * @author Emanuel Rabina
 */
public class AudFileStreaming extends AudFile implements Streaming {

	/**
	 * Constructor, assigns the filename to be the name of the file, and
	 * generates a new <code>FilePointer</code> using that filename.
	 * 
	 * @param name The file to be parsed.
	 */
	public AudFileStreaming(String name) {

		super(name);
	}

	/**
	 * Constructor, takes a name for the name of the file and the
	 * <code>FilePointer</code> as the location of the file.  Body processing
	 * only occurs on larger AUD files, such as music tracks.
	 * 
	 * @param name The name of this file.
	 * @param file Pointer to the file to parse.
	 */
	public AudFileStreaming(String name, FilePointer file) {

		super(name, file);
	}

	/**
	 * @inheritDoc
	 */
	public DataStream getPipe() {

		final DataStream pipe = new DataStream(1);
		pipe.setAlgorithm(new Runnable() {
			public void run() {

				Thread.currentThread().setName("AUD file decoding thread - " + name);
				byte pcmtype = audheader.type;
				int[] update = { 0, 0 };

				// Decode all remaining chunks
				while (file.remaining() > 0) {

					// Sound chunk
					AudChunk audchunk = new AudChunk(file);
					ByteBuffer sourcebytes = file.readBytes(audchunk.filesize);
					ByteBuffer destbytes   = BufferUtility.newByteBuffer(audchunk.datasize);

					// Select the proper decompression scheme and decompress
					if (pcmtype == TYPE_WS_ADPCM) {
						CNCCodec.decompress8bitWSADPCM(sourcebytes, destbytes);
					}
					else if (pcmtype == TYPE_IMA_ADPCM) {
						CNCCodec.decompress16bitIMAADPCM(sourcebytes, destbytes, update);
					}

					// Push decoded chunk down the pipe
					pipe.getSink().write(new ByteBuffer[]{ destbytes });
				}
				pipe.getSink().finish();
			}
		});
		return pipe;
	}

	/**
	 * @inheritDoc
	 */
	public ByteBuffer getSound() {

		return null;
	}

	/**
	 * @inheritDoc
	 */
	public void readFromFile() {

		// Init header, leave file at end of header for streaming file
		audheader = new AudHeader(file);
	}
}
