
// ===========================================
// Scanner's Java - Bitmap file implementation
// ===========================================

package redhorizon.filetypes.bmp;

import redhorizon.filetypes.FileImplementation;
import redhorizon.filetypes.ImageFile;
import redhorizon.filetypes.UnsupportedFileTypeException;
import redhorizon.utilities.BufferUtility;
import redhorizon.utilities.FilePointer;

import java.awt.image.BufferedImage;
import java.awt.image.DataBufferByte;
import java.io.ByteArrayInputStream;
import java.nio.ByteBuffer;
import javax.imageio.ImageIO;

/**
 * Implementation of the Windows bitmap image filetype.  The supported bitmap
 * types are limited to what is currently supported in Java.
 * 
 * @author Emanuel Rabina
 */
public class BmpFile extends FileImplementation implements ImageFile {

	private ByteBuffer bmpimage;
	private int width;
	private int height;

	/**
	 * Constructor, takes a file name and uses that to generate a new
	 * <code>FilePointer</code>.
	 * 
	 * @param name Name of the BMP file.
	 */
	public BmpFile(String name) {

		this(name, new FilePointer(name));
	}

	/**
	 * Constructor, takes the name of the file and uses the given
	 * <code>FilePointer</code> to parse the file.
	 * 
	 * @param name Name of the BMP file.
	 * @param file <code>FilePointer</code> to the BMP file.
	 */
	public BmpFile(String name, FilePointer file) {

		super(name, file);
	}

	/**
	 * @inheritDoc
	 */
	public int format() {

		return FORMAT_RGB;
	}

	/**
	 * @inheritDoc
	 */
	public ByteBuffer getImage() {

		return bmpimage;
	}

	/**
	 * @inheritDoc
	 */
	public int height() {

		return height;
	}

	/**
	 * @inheritDoc
	 */
	public void readFromFile() throws UnsupportedFileTypeException {

		// Convert FilePointer to InputStream for ImageIO
		ByteBuffer inputbytes = file.readBytes(file.size());
		ByteArrayInputStream inputstream = new ByteArrayInputStream(inputbytes.array());
		BufferedImage image = ImageIO.read(inputstream);

		// Fill bitmap attributes
		width  = image.getWidth();
		height = image.getHeight();

		// Switch R->B and B->R in BGR bitmaps
		byte[] bitmapbytes = ((DataBufferByte)image.getData().getDataBuffer()).getData();
		for (int i = 0; i < bitmapbytes.length; i += 3) {
			byte r2b = bitmapbytes[i];
			byte b2r = bitmapbytes[i + 2];
			bitmapbytes[i] = b2r;
			bitmapbytes[i + 2] = r2b;
		}

		// Save bitmap image
		bmpimage = BufferUtility.wrapByteBuffer(bitmapbytes);
	}

	/**
	 * @inheritDoc
	 */
	public int width() {

		return width;
	}
}
