
// ==================================================
// Scanner's Java - Representation of a settings file
// ==================================================

package redhorizon.filetypes.ini;

import redhorizon.filetypes.FileImplementation;
import redhorizon.filetypes.MapFile;
import redhorizon.filetypes.MissionFile;
import redhorizon.filetypes.SettingsFile;
import redhorizon.misc.CNCGameTypes;
import redhorizon.utilities.FilePointer;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

/**
 * Representation of an entire .ini file.  This file only contains an ordered
 * list of the sections which are found in the file, but by digging deeper via
 * the {@link IniSection}, the original parameters can also be retrieved.
 * 
 * @author Emanuel Rabina
 */
public class IniFile extends FileImplementation implements MapFile, MissionFile, SettingsFile {

	private ArrayList<Object> inivalues;
	private HashMap<String, Integer> sections;
	private int lastline;

	/**
	 * Constructor, takes a name for the name of the file, then generates a
	 * <code>FilePointer</code> to the file with that name.
	 * 
	 * @param name The name of the INI file.
	 */
	public IniFile(String name) {

		this(name, new FilePointer(name));
	}

	/**
	 * Constructor, takes a name for the name of the file, and the given
	 * <code>FilePointer</code>, then parses the information from the file using
	 * the pointer.
	 * 
	 * @param name The name of this file.
	 * @param file <code>FilePointer</code> of the file to parse.
	 */
	public IniFile(String name, FilePointer file) {

		super(name, file);
	}

	/**
	 * @inheritDoc
	 */
	public Map<String, String> getSection(String section) {

		int startindex = sections.get(section);
		HashMap<String, String> params = new HashMap<String, String>();

		// Run through the values for the section
		for (int i = startindex + 1; i < inivalues.size(); i++) {
			Object line = inivalues.get(i);

			// Skip non-IniParameters
			if (line instanceof String) {
				continue;
			}
			else if (line instanceof IniSection) {
				break;
			}

			IniParameter param = (IniParameter)line;
			params.put(param.name, param.value);
		}

		return params;
	}

	/**
	 * @inheritDoc
	 */
	public Map<String,Map<String,String>> getSections() {

		HashMap<String, Map<String, String>> allsections = new HashMap<String, Map<String, String>>();
		for (String sectionname: sections.keySet()) {
			allsections.put(sectionname, getSection(sectionname));
		}
		return allsections;
	}

	/**
	 * @inheritDoc
	 */
	public String getValue(String section, String parameter) {

		int lineno = sections.get(section);
		String result = null;

		for (int i = lineno + 1; true; i++) {
			Object line = inivalues.get(i);

			// Break if next section encountered
			if (line instanceof IniSection) {
				break;
			}
			// Test the parameter for equality
			else if (line instanceof IniParameter) {
				IniParameter param = (IniParameter)line;

				if (param.name.equals(parameter)) {
					result = ((IniParameter)line).value;
					break;
				}
			}
		}
		return result;
	}

	/**
	 * @inheritDoc
	 */
	public boolean hasSection(String section) {

		return sections.containsKey(section);
	}

	/**
	 * @inheritDoc
	 */
	public String missionName() {

		return getValue("[Basic]", "Name");
	}

	/**
	 * @inheritDoc
	 */
	public void readFromFile() {

		inivalues = new ArrayList<Object>();
		sections  = new HashMap<String, Integer>();

		// Start reading lines
		for ( ; file.remaining() > 0; lastline++) {
			String line = file.readLine().trim();

			// Include comments and empty lines
			if (line.length() == 0 || line.trim().charAt(0) == ';') {
				inivalues.add(line);
			}

			// Section headers
			else if (line.charAt(0) == '[') {
				IniSection section = new IniSection(line);
				inivalues.add(section);
				sections.put(section.toString(), lastline);
			}

			// Paramter types
			else {
				inivalues.add(new IniParameter(line));
			}
		}
	}

	/**
	 * @inheritDoc
	 */
	public CNCGameTypes targetGameType() {

		CNCGameTypes typeval = null;
		for (Object inivalue: inivalues) {

			if (inivalue instanceof IniParameter) {
				IniParameter iniparam = (IniParameter)inivalue;

				if (iniparam.name.equals("NewINIFormat")) {
					switch (Integer.parseInt(iniparam.value)) {
					case 1:
						typeval = CNCGameTypes.TIBERIUM_DAWN;
						break;
					case 3:
						typeval = CNCGameTypes.RED_ALERT;
						break;
					}
					break;
				}
			}
		}
		return typeval;
	}

	/**
	 * Returns a <code>String</code> representation of this INI file.
	 * 
	 * @return <code>String</code> representation of this file.
	 */
	public String toString() {

		String result = "";
		for (Object line: inivalues) {
			result += line + "\r\n";
		}
		return result;
	}
}
