
// =========================================================
// Scanner's Java - Representation of the Red Alert MIX file
// =========================================================

package redhorizon.filetypes.mix;

import redhorizon.filetypes.ArchiveFile;
import redhorizon.filetypes.FileImplementation;
import redhorizon.utilities.Blowfish;
import redhorizon.utilities.BufferUtility;
import redhorizon.utilities.FilePointer;
import redhorizon.utilities.WSkey;

import java.nio.ByteBuffer;

/**
 * Implementation of a Red Alert MIX file.  The MIX format is a file package,
 * much like a ZIP file, but without compression.  Documentation on this file
 * format, and other Red Alert file formats, can be found here:
 * <a href="http://xccu.sourceforge.net/">http://xccu.sourceforge.net/</a>
 * 
 * @author Emanuel Rabina
 */
public class MixFile extends FileImplementation implements ArchiveFile {

	private static final int FLAG_CHECKSUM  = 0x00010000;
	private static final int FLAG_ENCRYPTED = 0x00020000;

	private MixHeader mixheader;
	private boolean checksum;
	private boolean encrypted;

	/**
	 * Constructor, takes a name for the file to load and then generates a
	 * <code>FilePointer</code> from the name (assumes the name is a file on the
	 * disk).
	 * 
	 * @param name The name of the mix file.
	 */
	public MixFile(String name) {

		this(name, new FilePointer(name));
	}

	/**
	 * Constructor, takes a name for the name of the file, and parses the file
	 * pointed-to by the given <code>FilePointer</code>.  Body information is
	 * never loaded for MIX files, only mixheader info.
	 * 
	 * @param name The name of the mix file.
	 * @param file Pointer to the file to parse.
	 */
	public MixFile(String name, FilePointer file) {

		super(name, file);
	}

	/**
	 * Calculates an ID for a <code>MixRecord</code> given the original file
	 * name for the entry to which it is referring to.
	 * 
	 * @param name The original filename of the item in the MIX body.
	 * @return The ID of the entry from the filename.
	 */
	private static int calculateID(String name) {

		int i = 0;
		int id = 0;
		name = name.toUpperCase();
		while (i < name.length()) {
			int a = 0;
			for (int j = 0; j < 4; j++) {
				a >>>= 8;
				if (i < name.length()) {
					a += name.charAt(i) << 24;
				}
				i++;
			}
			id = (id << 1 | id >>> 31) + a;
		}
		return id;
	}

	/**
	 * @inheritDoc
	 */
	public FilePointer getItem(String itemname) {

		// Locate the record
		MixRecord record = getRecord(itemname);

		// Generate a new FilePointer, starting at the location of the item
		int offset = record.offset + headerSize();
		int length = record.length;

		return new FilePointer(name, offset, offset + length);
	}

	/**
	 * Returns a record for an item in the MIX file, instead of the item itself.
	 * Uses a binary search algorithm to locate the record.
	 * 
	 * @param recordname Name of the item and the record.
	 * @return <code>MixRecord</code> object of the record for the item.
	 */
	private MixRecord getRecord(String recordname) {

		int itemid = calculateID(recordname);

		// Binary search for the record with the calculated value
		MixRecord record = null;
		int lo = 0;
		int hi = mixheader.index.length - 1;

		while (lo <= hi) {
			int mid = (lo + hi) >> 1;
			int midval = mixheader.index[mid].id;
			
			if (itemid < midval) {
				hi = mid - 1;
			}
			else if (itemid > midval) {
				lo = mid + 1;
			}
			else {
				record = mixheader.index[mid];
				break;
			}
		}

		return record;
	}

	/**
	 * @inheritDoc
	 */
	public boolean hasItem(String itemname) {

		return (getRecord(itemname) != null);
	}

	/**
	 * Returns the size of the mixheader portion of the MIX file.  Used for
	 * properly setting the offset and locating the start of the MIX body.
	 * 
	 * @return Size, in bytes, of the mixheader.
	 */
	private int headerSize() {

		int flag = checksum || encrypted ? 4 : 0;
		int encryption = encrypted ? 80 : 0;
		int header = encrypted ? 8 : 6;
		int index = encrypted ? (int)Math.ceil((numFiles() * 12) / 8.0) * 8 : numFiles() * 12;

		return flag + header + encryption + index;
	}

	/**
	 * Returns the number of files inside the MIX file.
	 * 
	 * @return Number of files within.
	 */
	private int numFiles() {

		return mixheader.numfiles & 0xffff;
	}

	/**
	 * @inheritDoc
	 */
	public void readFromFile() {

		// Find out if this file has a checksum/encryption
		int flag = file.readBytes(4).getInt();
		checksum  = (flag & FLAG_CHECKSUM)  != 0;
		encrypted = (flag & FLAG_ENCRYPTED) != 0;

		// If encrypted, decrypt the mixheader and index
		if (encrypted) {

			// Perform the public -> private/Blowfish key function
			ByteBuffer keysource = file.readBytes(80);
			ByteBuffer key = BufferUtility.newByteBuffer(56);
			WSkey.getBlowfishKey(keysource, key);
			Blowfish.setKey(key, 56);

			// Decrypt the mixheader
			ByteBuffer headerbytes = file.readBytes(8);
			Blowfish.decrypt(headerbytes, headerbytes, 8);
			mixheader = new MixHeader(headerbytes);

			// Now figure-out how many more on 8-byte blocks (+2) to decrypt
			int numrecords = numFiles();
			int numblocks = (int)Math.ceil((12 * numrecords) / 8.0);
			int numbytes = numblocks * 8;

			ByteBuffer encryptedbytes = file.readBytes(numbytes);
			Blowfish.decrypt(encryptedbytes, encryptedbytes, numbytes);

			ByteBuffer recordsbytes = BufferUtility.newByteBuffer(numbytes + 2);
			recordsbytes.put(headerbytes).put(encryptedbytes).rewind();

			// Take all the decrypted data and turn them into the index records
			for (int i = 0; i < numrecords; i++) {
				MixRecord mixrecord = new MixRecord(recordsbytes);
				mixheader.addRecord(mixrecord);
			}
		}
		// If not encrypted, just read the straight data
		else {

			// Read the mixheader
			file.rewind();
			ByteBuffer headerbytes = file.readBytes(6);
			mixheader = new MixHeader(headerbytes);

			// Now figure-out how much more of the file is the index
			int numrecords = numFiles();
			int numblocks = 12 * numrecords;
			int numbytes = numblocks * 8;

			ByteBuffer recordsbytes = file.readBytes(numbytes);

			// Take all the data and turn it into the index records
			for (int i = 0; i < numrecords; i++) {
				MixRecord mixrecord = new MixRecord(recordsbytes);
				mixheader.addRecord(mixrecord);
			}
		}
	}
}
