
// =========================================================
// Scanner's Java - Representation of a record in a PAL file
// =========================================================

package redhorizon.filetypes.pal;

import redhorizon.utilities.FilePointer;

import java.nio.ByteBuffer;

/**
 * Representation of a record in the PAL file.  This is a colour entry in your
 * standard RGB format.<br>
 * <br>
 * Unlike most of the other classes in the {@link redhorizon.filetypes} package,
 * this class allows itself to be subclassed, as many of the older formats from
 * the C&C games rely on internal palettes, of which most can be described with
 * this class.
 * 
 * @author Emanuel Rabina
 */
public class PalColour {

	private byte red;
	private byte green;
	private byte blue;

	/**
	 * Constructor, assigns the RGB components of the colour from 3 bytes taken
	 * from the file.  Assumes the values to be read are in old VGA colour form.
	 * 
	 * @param file <code>FilePointer</code> to the file.
	 */
	protected PalColour(FilePointer file) {

		this(file, true);
	}

	/**
	 * Constructor, assigns RGB components from 3 bytes taken from the
	 * <code>ByteBuffer</code>.  Can specify if the format of the bytes is in
	 * old VQA colour (<code>true</code>) or the standard 0-255 colour range
	 * (<code>false</code>).
	 * 
	 * @param file	  <code>FilePointer</code> to the file.
	 * @param vgaform Specifies whether or not the RGB components are in the old
	 * 				  VQA form.
	 */
	protected PalColour(FilePointer file, boolean vgaform) {

		this(file.readBytes(3), vgaform);
	}

	/**
	 * Constructor, assigns the RGB components of the colour from 3 bytes taken
	 * from the given <code>ByteBuffer</code>.  Assumes the values to be read
	 * are in old VGA colour form.
	 * 
	 * @param bytes	<code>ByteBuffer</code> of colour data.
	 */
	protected PalColour(ByteBuffer bytes) {

		this(bytes, true);
	}

	/**
	 * Constructor, assigns the RGB components of the colour from 3 bytes taken
	 * from the given <code>ByteBuffer</code>.  Can specify if the format of the
	 * bytes is in old VQA colour (<code>true</code>) or the standard 0-255
	 * colour range (<code>false</code>).
	 * 
	 * @param bytes	  <code>ByteBuffer</code> of colour data.
	 * @param vgaform Specifies whether or not the RGB components are in the old
	 * 				  VQA form.
	 */
	protected PalColour(ByteBuffer bytes, boolean vgaform) {

		// Raw byte values are in 6-bit VGA format
		if (vgaform) {
			red   = (byte)(bytes.get() & 0x3f);
			green = (byte)(bytes.get() & 0x3f);
			blue  = (byte)(bytes.get() & 0x3f);

			red   <<= 2;
			green <<= 2;
			blue  <<= 2;
		}
		// New 0-255 range
		else {
			red   = bytes.get();
			green = bytes.get();
			blue  = bytes.get();
		}
	}

	/**
	 * Constructor, provides the raw red/green/blue bytes.
	 * 
	 * @param red	Red component.
	 * @param green Green component.
	 * @param blue	Blue component.
	 */
	protected PalColour(byte red, byte green, byte blue) {

		this.red   = red;
		this.green = green;
		this.blue  = blue;
	}

	/**
	 * Returns the red component of this colour.
	 * 
	 * @return The R in RGB.
	 */
	public byte getRed() {

		return red;
	}

	/**
	 * Returns the green component of this colour.
	 * 
	 * @return The G in RGB.
	 */
	public byte getGreen() {

		return green;
	}

	/**
	 * Returns the blue component of this colour.
	 * 
	 * @return The B in RGB.
	 */
	public byte getBlue() {

		return blue;
	}

	/**
	 * Returns the RGB components as an array.  Override this method in custom
	 * palette colour implementations so that the image utilities palette
	 * application method uses the correct colour ordering/parts.
	 * 
	 * @return <code>byte[]</code> with the 0th, 1st, and 2nd representing the
	 * 		   red, green, and blue parts respectively.
	 */
	public byte[] toArray() {

		return new byte[]{ red, green, blue };
	}
}
