
// =========================================================
// Scanner's Java - Representation of the Red Alert PAL file
// =========================================================

package redhorizon.filetypes.pal;

import redhorizon.filetypes.FileImplementation;
import redhorizon.filetypes.PaletteAlpha;
import redhorizon.filetypes.PaletteShift;
import redhorizon.filetypes.PaletteFile;
import redhorizon.utilities.FilePointer;

import java.util.HashMap;
import java.util.Map;

/**
 * Implementation of the Red Alert PAL file.  Essentially just an array of the
 * 256 colours that Red Alert uses in each of the palettes.
 * 
 * @author Emanuel Rabina
 */
public class PalFile extends FileImplementation implements PaletteFile {

	private PalColour[] palcolours;

	// Transparency on a per-index basis
	private Map<Integer,byte[]> alpha;
	private String alphaused;

	// Shift palette, used to remap index entries to side-specific colours.
	private Map<Byte,Byte> shift;
	private String shiftused;

	/**
	 * Constructor, uses the given name as the name of the file, then generates
	 * a <code>FilePointer</code> using that name.
	 * 
	 * @param name The name of the SHP file.
	 */
	public PalFile(String name) {

		this(name, new FilePointer(name));
	}

	/**
	 * Constructor, uses the given name for the file, then parses the file using
	 * the given <code>FilePointer</code>.
	 * 
	 * @param name The name of this file.
	 * @param file <code>FilePointer</code> to the file.
	 */
	public PalFile(String name, FilePointer file) {

		super(name, file);
	}

	/**
	 * @inheritDoc
	 */
	public void applyAlpha(PaletteAlpha palalpha) {

		Map<Integer,int[]> alphamap = palalpha.getAlphaMap();
		HashMap<Integer,byte[]> result = new HashMap<Integer,byte[]>();

		// Keep only non -1 values
		for (int index: alphamap.keySet()) {
			byte[] replacement = new byte[4];
			int[] alphacheck = alphamap.get(index);

			// Replace incoming value with new or current
			for (int i = 0; i < 4; i++) {
				if (alphacheck[i] == -1) {
					alphacheck[i] = getRGBA(index)[i];
				}
				replacement[i] = (byte)alphacheck[i];
			}
			result.put(index, replacement);
		}

		alpha     = result;
		alphaused = palalpha.getAlphaName();
	}

	/**
	 * @inheritDoc
	 */
	public void applyShift(PaletteShift palshift) {

		shift     = palshift.getShiftMap();
		shiftused = palshift.getShiftName();
	}

	/**
	 * @inheritDoc
	 */
	public String getFilename() {

		return name + (shiftused != null ? "_" + shiftused : "") + (alphaused != null ? "_" + alphaused : "");
	}

	/**
	 * @inheritDoc
	 */
	public byte[] getRGB(int index) {

		return new byte[] {
			palcolours[shift != null && shift.containsKey(index) ?
				shift.get(index) : index].getRed(),
			palcolours[shift != null && shift.containsKey(index) ?
				shift.get(index) : index].getGreen(),
			palcolours[shift != null && shift.containsKey(index) ?
				shift.get(index) : index].getBlue(),
		};
	}

	/**
	 * @inheritDoc
	 */
	public byte[] getRGBA(int index) {

		byte[] rgba;
		if (alpha != null && alpha.containsKey(index)) {
			rgba = alpha.get(index);
		}
		else {
			byte[] rgb = getRGB(index);
			rgba = new byte[]{ rgb[0], rgb[1], rgb[2], (byte)255 };
		}
		return rgba;
	}

	/**
	 * @inheritDoc
	 */
	public void readFromFile() {

		// Fills the file palette with the colours from the palette file.
		palcolours = new PalColour[256];
		for (int i = 0; i < 256; i++) {
			palcolours[i] = new PalColour(file);
		}
	}
}
