
// ========================================
// Scanner's Java - PCX file representation
// ========================================

package redhorizon.filetypes.pcx;

import redhorizon.filetypes.FileImplementation;
import redhorizon.filetypes.ImageFile;
import redhorizon.utilities.BufferUtility;
import redhorizon.utilities.CNCCodec;
import redhorizon.utilities.FilePointer;
import redhorizon.utilities.ImageUtility;

import java.nio.ByteBuffer;

/**
 * Implementation of the PCX file format.  PCX files are used for the higher
 * resolution (640x400) still images used in Red Alert and Tiberium Dawn.<br>
 * <br>
 * PCX files can come in many flavours (8-bit, 16-bit, 24-bit, no palette, etc),
 * but for the purpose of Red Horizon, PCX files will be of the type used in the
 * old RA & TD games: a 256-colour file with an internal palette located at the
 * tail of the file.  Adding support for other sorts of PCX files may be done in
 * future.
 * 
 * @author Emanuel Rabina.
 */
public class PcxFile extends FileImplementation implements ImageFile {

	private static final byte ENCODING_RLE = 1;

	private PcxHeader pcxheader;
	private PcxColour[] pcxpalette;
	private ByteBuffer pcximage;

	/**
	 * Constructor, takes a file name and uses that to generate a new
	 * <code>FilePointer</code>.
	 * 
	 * @param name Name of the PCX file.
	 */
	public PcxFile(String name) {

		this(name, new FilePointer(name));
	}

	/**
	 * Constructor, takes the name of the file and uses the given
	 * <code>FilePointer</code> to parse the file.
	 * 
	 * @param name Name of the PCX file.
	 * @param file <code>FilePointer</code> to the PCX file.
	 */
	public PcxFile(String name, FilePointer file) {

		super(name, file);
	}

	/**
	 * Returns the number of bytes to reserve for a scan line.
	 * 
	 * @return Bytes per scan line.
	 */
	private int bytesPerLine() {

		return pcxheader.bytesperline & 0xffff;
	}

	/**
	 * @inheritDoc
	 */
	public int format() {

		return FORMAT_RGB;
	}

	/**
	 * @inheritDoc
	 */
	public ByteBuffer getImage() {

		return pcximage;
	}

	/**
	 * @inheritDoc
	 */
	public int height() {

		return pcxheader.ymax - pcxheader.ymin + 1;
	}

	/**
	 * @inheritDoc
	 */
	public void readFromFile() {

		// Init header & final image buffer
		pcxheader  = new PcxHeader(file);
		pcxpalette = new PcxColour[256];

		// Assign palette (from tail of file)
		file.seek(file.size() - 768);
		for (int i = 0; i < 256; i++) {
			pcxpalette[i] = new PcxColour(file, false);
		}
		file.seek(128);

		// Decode PCX run-length encoded image data
		pcximage  = BufferUtility.newByteBuffer(width() * height() * format());
		ByteBuffer sourcebytes = file.readBytes(file.size() - 769 - 128);
		ByteBuffer framebytes;
		ByteBuffer cullbytes;

		// Decode RLE image
		if (pcxheader.encoding == ENCODING_RLE) {
			int framesize = bytesPerLine() * height();
			framebytes = BufferUtility.newByteBuffer(framesize);
			CNCCodec.decodeRLE67(sourcebytes, framebytes);
		}
		else {
			framebytes = sourcebytes;
		}

		// Cull the image to the appropriate width/height dimensions (for when
		// scanlines extend beyond the image borders)
		if (bytesPerLine() != width()) {
			int cullsize = width() * height();
			cullbytes = BufferUtility.newByteBuffer(cullsize);

			for (int i = 0; cullbytes.hasRemaining(); i++) {
				int position = i % bytesPerLine();
				if (position >= pcxheader.xmin && position <= pcxheader.xmax) {
					cullbytes.put(framebytes.get());
				}
			}
		}
		else {
			cullbytes = framebytes;
		}

		// Apply the internal palette
		ImageUtility.applyPalette(cullbytes, pcximage, pcxpalette);
	}

	/**
	 * @inheritDoc
	 */
	public int width() {

		return pcxheader.xmax - pcxheader.xmin + 1;
	}
}
