
// ===============================================================
// Scanner's Java - Portable Network Graphics (PNG) implementation
// ===============================================================

package redhorizon.filetypes.png;

import redhorizon.filetypes.FileImplementation;
import redhorizon.filetypes.ImageFile;
import redhorizon.filetypes.ImagesFile;
import redhorizon.filetypes.Paletted;
import redhorizon.filetypes.pal.PalFile;
import redhorizon.filetypes.shp.ShpFileDune2;
import redhorizon.misc.CNCAlphaMask;
import redhorizon.strings.Common_Filetypes;
import redhorizon.strings.ProgramStrings;
import redhorizon.utilities.BufferUtility;
import redhorizon.utilities.FilePointer;
import redhorizon.utilities.ImageUtility;

import java.awt.image.BufferedImage;
import java.awt.image.DataBufferByte;
import java.awt.image.IndexColorModel;
import java.io.ByteArrayInputStream;
import java.io.FileOutputStream;
import java.nio.ByteBuffer;

import javax.imageio.ImageIO;

/**
 * Implementation of the Portable Network Graphics (PNG) image file type.  The
 * supported PNG types are a subset of what is currently supported in Java.
 * 
 * @author Emanuel Rabina
 */
public class PngFile extends FileImplementation implements ImageFile {

	private ByteBuffer pngimage;
	private int format;
	private int width;
	private int height;
	private boolean paletted;
	private PngColour[] palette;

	// Can create a PNG file from the following types
	private ImagesFile imagesfile;
	private PalFile palfile;

	/**
	 * Constructor, takes a file name and uses that to generate a new
	 * <code>FilePointer</code>.
	 * 
	 * @param name Name of the PNG file.
	 */
	public PngFile(String name) {

		this(name, new FilePointer(name));
	}

	/**
	 * Constructor, takes the name of the file and uses the given
	 * <code>FilePointer</code> to parse the file.
	 * 
	 * @param name Name of the PNG file.
	 * @param file <code>FilePointer</code> to the PNG file.
	 */
	public PngFile(String name, FilePointer file) {

		super(name, file);
	}

	/**
	 * Constructor, allows the construction of a new PNG file from the given
	 * <code>ImagesFile</code> implementation, and the accompanying parameters.
	 * 
	 * @param name		 The name of this file.
	 * @param imagesfile Pre-read <code>ImagesFile</code> from which to extract
	 * 					 the data from.
	 * @param params	 Additional parameters for a PNG file conversion.
	 * 					 Supported parameters are:<br>
	 * 					  - 0th = filename of a palette to use
	 */
	private PngFile(String name, ImagesFile imagesfile, String... params) {

		this(name, new FilePointer(name, false));
		this.imagesfile = imagesfile;

		// Load palette from 0th parameter
		palfile = new PalFile(params[0]);
		palfile.readFromFile();
	}

	/**
	 * Constructor, allows the construction of a new PNG file from the Dune 2
	 * SHP file implementation, and the accompanying parameters.
	 * 
	 * @param name		The name of this file.
	 * @param shpfiled2 Pre-read Dune 2 SHP file from which to extract the data
	 * 					from.
	 * @param params	Additional parameters for a PNG file conversion.
	 * 					Supported parameters are:<br>
	 * 					 - 0th = filename of a palette to use
	 */
	public PngFile(String name, ShpFileDune2 shpfiled2, String... params) {

		this(name, (ImagesFile)shpfiled2, params);
	}

	/**
	 * @inheritDoc
	 */
	public int format() {

		return format;
	}

	/**
	 * @inheritDoc
	 */
	public ByteBuffer getImage() {

		// Apply RGB palette if paletted
		if (paletted) {
			ByteBuffer rgbimage = BufferUtility.newByteBuffer(width * height * format);
			ImageUtility.applyPalette(pngimage, rgbimage, palette);
			return rgbimage;
		}

		return pngimage;
	}

	/**
	 * Returns the raw indexed data which constructed this image.  Only
	 * available if {@link #isIndexed()} returns <code>true</code>.
	 * 
	 * @return The indexed image data, or <code>null</code> if not indexed.
	 */
	public ByteBuffer getRawImage() {

		return paletted ? pngimage : null;
	}

	/**
	 * @inheritDoc
	 */
	public int height() {

		return height;
	}

	/**
	 * Indicates that this PNG image came from indexed (paletted) data.  If it
	 * is indexed, then the raw indexed data can be obtained using
	 * {@link #getRawImage()}.
	 * 
	 * @return <code>true</code> if the PNG was built from indexed data,
	 * 		   <code>false</code> otherwise.
	 */
	public boolean isIndexed() {

		return paletted;
	}

	/**
	 * @inheritDoc
	 */
	public void readFromFile() {

		// Convert FilePointer to InputStream for ImageIO
		ByteBuffer inputbytes = file.readBytes(file.size());
		ByteArrayInputStream inputstream = new ByteArrayInputStream(inputbytes.array());
		BufferedImage image = ImageIO.read(inputstream);

		// Fill image attributes and image data
		width    = image.getWidth();
		height   = image.getHeight();
		format   = image.getColorModel().getNumComponents();
		pngimage = BufferUtility.wrapByteBuffer(((DataBufferByte)image.getData().getDataBuffer()).getData());

		// Indicate that this PNG came from indexed data
		paletted = image.getType() == BufferedImage.TYPE_BYTE_INDEXED;

		// Complete RGB palfile if paletted
		if (paletted) {
			int[] rgb = new int[256];
			((IndexColorModel)image.getColorModel()).getRGBs(rgb);

			palette = new PngColour[256];
			for (int i = 0; i < 256; i++) {
				byte red   = (byte)((rgb[i] >> 16) & 0xff);
				byte green = (byte)((rgb[i] >> 8) & 0xff);
				byte blue  = (byte) (rgb[i] & 0xff);
				if (format == FORMAT_RGBA) {
					byte alpha = (byte)((rgb[i] >> 24) & 0xff);
					palette[i] = new PngColour(red, green, blue, alpha);
				}
				else {
					palette[i] = new PngColour(red, green, blue);
				}
			}
		}
	}

	/**
	 * @inheritDoc
	 */
	public void saveToFile() {

		// Get the maximum dimensions
		int numimages = imagesfile.numImages();
		int maximgwidth  = 0;
		int maximgheight = 0;
		for (int i = 0; i < numimages; i++) {
			if (maximgwidth < imagesfile.width(i)) {
				maximgwidth = imagesfile.width(i);
			}
			if (maximgheight < imagesfile.height(i)) {
				maximgheight = imagesfile.height(i);
			}
		}
		width  = numimages <= 16 ? maximgwidth * numimages : maximgwidth * 16;
		height = maximgheight * (int)Math.ceil(numimages / 16d);

		// Create paletted PNG
		if (imagesfile instanceof Paletted) {
			format = FORMAT_RGBA;
			paletted = true;

			// Build the IndexColorModel (Java palette), use alpha
			byte[] reds   = new byte[256];
			byte[] greens = new byte[256];
			byte[] blues  = new byte[256];
			byte[] alphas = new byte[256];

			palfile.applyAlpha(CNCAlphaMask.INDEX0_ONLY);

			for (int i = 0; i < 256; i++) {
				byte[] rgba = palfile.getRGBA(i);
				reds[i]   = rgba[0];
				greens[i] = rgba[1];
				blues[i]  = rgba[2];
				alphas[i] = rgba[3];
			}
			IndexColorModel colormodel = new IndexColorModel(8, 256, reds, greens, blues, alphas);

			// Create new BufferedImage using the IndexColorModel, width, height
			BufferedImage outputimg = new BufferedImage(width, height, BufferedImage.TYPE_BYTE_INDEXED, colormodel);
			byte[] outputimgdata = ((DataBufferByte)outputimg.getRaster().getDataBuffer()).getData();

			// Write image data from the source file into the BufferedImage
			ByteBuffer[] sourcebytes = imagesfile.getImages();
			for (int i = 0, row = 0, col = 0; i < numimages; ) {
				ByteBuffer sourceimgbytes = sourcebytes[i];
				int imgwidth  = imagesfile.width(i);
				int imgheight = imagesfile.height(i);

				// Copy from source into image data
				for (int y = 0; y < imgheight; y++) {
					for (int x = 0; x < imgwidth; x++) {
						outputimgdata[row + (y * width) + (x + col)] = sourceimgbytes.get((y * imgwidth) + x);
					}
				}

				// Adjust current data co-ordinates in array
				i++;
				if (i % 16 == 0) {
					row += width * maximgheight;
				}
				col = (i % 16) * maximgwidth;
			}

			// Write to file
			ImageIO.write(outputimg, "PNG", new FileOutputStream(file.getFileDescriptor()));
		}

		// Create 32-bit PNG
		else {
			super.saveToFile();
		}
	}

	/**
	 * Returns some information on this PNG file.
	 * 
	 * @return PNG file info.
	 */
	public String toString() {

		return name + " (" + getClass().getSimpleName() + ")" +
			"\n  " + ProgramStrings.getText(Common_Filetypes.IMAGE_DIMX) + ": " + width +
			"\n  " + ProgramStrings.getText(Common_Filetypes.IMAGE_DIMY) + ": " + height +
			"\n  " + ProgramStrings.getText(Common_Filetypes.IMAGE_COLOURS) + ": " + (format * 8) + "-" +
					 ProgramStrings.getText(Common_Filetypes.IMAGE_COLOURS_BIT) + " " +
					 (paletted ? ProgramStrings.getText(Common_Filetypes.IMAGE_COLOURS_PALETTE) : "");
	}

	/**
	 * @inheritDoc
	 */
	public int width() {

		return width;
	}
}
