
// =============================================
// Scanner's Java - representation of a SHP file
// =============================================

package redhorizon.filetypes.shp;

import redhorizon.filetypes.AnimationFile;
import redhorizon.filetypes.FileImplementation;
import redhorizon.filetypes.ImagesFile;
import redhorizon.filetypes.PaletteFile;
import redhorizon.filetypes.Paletted;
import redhorizon.strings.Common_Filetypes;
import redhorizon.strings.ProgramStrings;
import redhorizon.utilities.BufferUtility;
import redhorizon.utilities.CNCCodec;
import redhorizon.utilities.FilePointer;
import redhorizon.utilities.ImageUtility;

import java.nio.ByteBuffer;

/**
 * Implementation of a Red Alert SHP file.  SHP files are the graphics format
 * for almost everything in the game.
 * 
 * @author Emanuel Rabina
 */
public class ShpFile extends FileImplementation implements ImagesFile, AnimationFile, Paletted {

	private static final byte FORMAT20 = (byte)0x20;
	private static final byte FORMAT40 = (byte)0x40;
	private static final byte FORMAT80 = (byte)0x80;

	private ShpFileHeader shpheader;
	private ByteBuffer[] shpimages;

	/**
	 * Constructor, uses the given name as the name of the file, then generates
	 * a <code>FilePointer</code> from that name.
	 * 
	 * @param name The name of the SHP file.
	 */
	public ShpFile(String name) {

		this(name, new FilePointer(name));
	}

	/**
	 * Constructor, takes a name and, using the <code>FilePointer</code>, parses
	 * the file into useable data.
	 * 
	 * @param name The name of this file.
	 * @param file <code>FilePointer</code> to the SHP file.
	 */
	public ShpFile(String name, FilePointer file) {

		super(name, file);
	}

	/**
	 * @inheritDoc
	 */
	public float adjustmentFactor() {

		return 1f;
	}

	/**
	 * @inheritDoc
	 */
	public ByteBuffer[] applyPalette(PaletteFile palette) {

		ByteBuffer[] rgbaimages = new ByteBuffer[numImages()];
		for (int i = 0; i < numImages(); i++) {
			rgbaimages[i] = BufferUtility.newByteBuffer(width(i) * height (i) * format());
		}
		ImageUtility.applyPalette(shpimages, rgbaimages, format(), palette);
		return rgbaimages;
	}

	/**
	 * @inheritDoc
	 */
	@SuppressWarnings("static-access")
	public int format() {

		return ImagesFile.FORMAT_RGBA;
	}

	/**
	 * @inheritDoc
	 */
	public float frameRate() {

		return 15f;
	}

	/**
	 * @inheritDoc
	 */
	public ByteBuffer[] getImages() {

		return shpimages;
	}

	/**
	 * @inheritDoc
	 */
	public int height() {

		return shpheader.height & 0xffff;
	}

	/**
	 * @inheritDoc
	 */
	public int height(int imagenum) {

		return height();
	}

	/**
	 * Returns the offset information associated with the given frame number.
	 * 
	 * @param image The image number which we're interested in.
	 * @return The offset information for that frame.
	 */
	private ShpImageOffset imageOffset(int image) {

		return shpheader.offsets[image];
	}

	/**
	 * @inheritDoc
	 */
	public int numImages() {

		return shpheader.numimages & 0xffff;
	}

	/**
	 * @inheritDoc
	 */
	public void readFromFile() {

		shpheader = new ShpFileHeader(file);

		// numImages() + 2 for the 0 offset and EOF pointer
		for (int i = 0; i < numImages() + 2; i++) {
			shpheader.addOffSet(new ShpImageOffset(file));
		}

		// Decompresses the raw SHP data into palette-index data
		shpimages = new ByteBuffer[numImages()];

		// Decompress every frame
		for (int i = 0; i < numImages(); i++) {

			// Information on the current image
			ShpImageOffset imageoffset = imageOffset(i);
			int sourceoffset = imageoffset.offset;
			byte imageformat = imageoffset.offsetformat;

			// Format conversion buffers
			int sourcelength = imageOffset(i + 1).offset - sourceoffset;
			ByteBuffer sourcebytes = file.readBytes(sourcelength);

			int destlength = width() * height();
			ByteBuffer destbytes = BufferUtility.newByteBuffer(destlength);

			// Format80 image
			if (imageformat == FORMAT80) {
				CNCCodec.decodeFormat80(sourcebytes, destbytes);
			}
			// Format40 image
			else if (imageformat == FORMAT40) {
				int refoffset = imageoffset.refoff;

				// Find out which offset this reference belongs to
				int j;
				for (j = 0; j < numImages(); j++) {
					if (refoffset == imageOffset(j).offset) {
						break;
					}
				}
				destbytes.put(shpimages[j]).rewind();
				shpimages[j].rewind();

				CNCCodec.decodeFormat40(sourcebytes, destbytes);
			}
			// Format20 image
			else if (imageformat == FORMAT20) {
				destbytes.put(shpimages[i - 1]).rewind();
				shpimages[i - 1].rewind();

				CNCCodec.decodeFormat20(sourcebytes, destbytes);
			}

			// Add the decompressed image to the image array
			shpimages[i] = destbytes;
		}
	}

	/**
	 * Returns some information on this SHP file.
	 * 
	 * @return SHP file info.
	 */
	public String toString() {

		// Check if widths/heights of this file vary
		boolean variedwidth = false;
		int initwidth = width(0);
		for (int i = 1; i < numImages(); i++) {
			if (initwidth != width(i)) {
				variedwidth = true;
				break;
			}
		}
		boolean variedheight = false;
		int initheight = height(0);
		for (int i = 1; i < numImages(); i++) {
			if (initheight != height(i)) {
				variedheight = true;
				break;
			}
		}

		return name + " (SHP file)" +
			"\n  " + ProgramStrings.getText(Common_Filetypes.IMAGE_NUM) + " " + numImages() +
			"\n  " + ProgramStrings.getText(Common_Filetypes.IMAGE_DIMX) + ": " +
					 (variedwidth ? ProgramStrings.getText(Common_Filetypes.IMAGE_DIM_VARIES) : initwidth) +
			"\n  " + ProgramStrings.getText(Common_Filetypes.IMAGE_DIMY) + ": " +
				 	 (variedheight ? ProgramStrings.getText(Common_Filetypes.IMAGE_DIM_VARIES) : initheight) +
			"\n  " + ProgramStrings.getText(Common_Filetypes.IMAGE_COLOURS) + ": 8-" +
				 	 ProgramStrings.getText(Common_Filetypes.IMAGE_COLOURS_BIT);
	}

	/**
	 * @inheritDoc
	 */
	public int width() {

		return shpheader.width & 0xffff;
	}

	/**
	 * @inheritDoc
	 */
	public int width(int imagenum) {

		return width();
	}
}
