
// ===============================================
// Scanner's Java - Dune 2 SHP file representation
// ===============================================

package redhorizon.filetypes.shp;

import redhorizon.filetypes.FileImplementation;
import redhorizon.filetypes.ImageFile;
import redhorizon.filetypes.ImagesFile;
import redhorizon.filetypes.PaletteFile;
import redhorizon.filetypes.Paletted;
import redhorizon.filetypes.png.PngFile;
import redhorizon.strings.Common_Filetypes;
import redhorizon.strings.ProgramStrings;
import redhorizon.utilities.BufferUtility;
import redhorizon.utilities.CNCCodec;
import redhorizon.utilities.FilePointer;
import redhorizon.utilities.ImageUtility;

import java.nio.ByteBuffer;

/**
 * Implementation of the Dune 2 SHP file, used for extracting information and
 * image data from the file.  This type supports the ability to save itself to a
 * file.<br>
 * <br>
 * Details on this filetype are not very common, so I've written comments
 * (visible in the source, not in this JavaDoc) on how the file is structured,
 * so that others it can help others understand how it works if ever they wish
 * to write programs to utilize the file.
 * 
 * @author Emanuel Rabina
 */

/* 
 * ======================
 * Dune 2 SHP file format
 * ======================
 * 
 * The Dune 2 SHP file, is a multiple image filetype, where each image can have
 * it's own set of dimensions.  The file is structured as follows:
 * 
 * File header:
 *   NumImages (2 bytes)    - the number of images in the file
 *   Offsets[NumImages + 1] - offset to the image header for an image.  The last
 *     (2 or 4 bytes each)    offset points to the end of the file.  The offsets
 *                            don't take into account the NumImages bytes at the
 *                            beginning, so add 2 bytes to the offset value to
 *                            get the actual position of an image header in the
 *                            file
 * 
 * The size of the offsets can be either 2 or 4 bytes.  There is no simple way
 * to determine which it will be, but checking the file's 4th byte to see if
 * it's 0, seems to be a commonly accepted practice:
 * 
 * eg: A 2-byte offset file: 01 00 06 00 EC 00 45 0A ...
 *     A 4-byte offset file: 01 00 08 00 00 00 EC 00 ...
 *                                       ^^
 * The marked byte will be 0 in 4-byte offset files, non 0 in 2-byte offset
 * files.
 * Lastly, like C&C SHP files, there is an extra offset, pointing to the end of
 * the file (or what would have been the position of another image header/data
 * pair).
 * 
 * Following the file header, are a series of image header & image data pairs.
 * The image header is structured as follows:
 * 
 * Image header:
 *   Compression (2 bytes) - flags to identify the types of compression applied
 *                           to the following image data
 *   Slices (1 byte)       - number of Format2 slices used to encode the image
 *                           data.  Often this is the same as the height of the
 *                           image
 *   Width (2 bytes)       - width of the image
 *   Height (1 byte)       - height of the image
 *   Filesize (2 bytes)    - size of the image data in the file
 *   Datasize (2 bytes)    - size of the image data when Format2
 *                           encoded/compressed
 * 
 * Regarding the compression flags, if only Format2 compression is used on the
 * following image data, the 2nd LSB is set to 1.  If the Format2 is then
 * compressed further with Format80, then all bits are set to 0.
 * In the XCC source code, there also seems to be a case for when the LSB is set
 * to 1, but I have never encountered it before.
 * 
 * Immediately after the image header is the image data.
 */
public class ShpFileDune2 extends FileImplementation implements ImagesFile, Paletted {

	// Read-from information
	private static final int IMAGE_OFFSET_2 = 2;
	private static final int IMAGE_OFFSET_4 = 4;

	private ShpFileHeaderDune2 shpfileheader;
	private ShpImageHeaderDune2[] shpimageheaders;
	private int shpoffsetsize;
	private ByteBuffer[] shpimages;

	// Save-to information
	private static final String PARAM_WIDTH   = "-w:";
	private static final String PARAM_HEIGHT  = "-h:";
	private static final String PARAM_NUMIMGS = "-n:";

	private static final int MAX_WIDTH   = 65535;
	private static final int MAX_HEIGHT  = 255;
	private static final int MAX_NUMIMGS = 65535;
	private static final int MAX_IMGSIZE = 65535;

	private ImageFile srcimgfile;
	private ImagesFile srcimgsfile;
	private int destwidth = -1;
	private int destheight = -1;
	private int destnumimgs = -1;

	/**
	 * Constructor, uses the given name as the name of the file, then generates
	 * a <code>FilePointer</code> from that name.
	 * 
	 * @param name The name of the SHP file.
	 */
	public ShpFileDune2(String name) {

		this(name, new FilePointer(name));
	}

	/**
	 * Constructor, takes a name and, using the <code>FilePointer</code>, parses
	 * the file into useable data.
	 * 
	 * @param name The name of this file.
	 * @param file <code>FilePointer</code> to the SHP file.
	 */
	public ShpFileDune2(String name, FilePointer file) {

		super(name, file);
	}

	/**
	 * Constructor, builds this file from an <code>ImageFile</code>
	 * implementation.
	 * 
	 * @param name		Name of this file.
	 * @paarm srcimgfile File to source data from.
	 * @param params	Additonal parameters.
	 */
	private ShpFileDune2(String name, ImageFile imagefile, String... params) {

		this(name, new FilePointer(name, false));
		srcimgfile = imagefile;

		for (String param: params) {
			if (param.startsWith(PARAM_WIDTH)) {
				destwidth = Integer.parseInt(param.substring(3));
			}
			else if (param.startsWith(PARAM_HEIGHT)) {
				destheight = Integer.parseInt(param.substring(3));
			}
			else if (param.startsWith(PARAM_NUMIMGS)) {
				destnumimgs = Integer.parseInt(param.substring(3));
			}
		}

		// Ensure each parameter was filled
		if (destwidth == -1 || destheight == -1 || destnumimgs == -1) {
			throw new IllegalArgumentException();
		}
	}

	/**
	 * Constructor, builds this filetype using the data from a
	 * <code>PngFile</code> type.
	 * 
	 * @param name	  The name of this file.
	 * @param pngfile Fully read PNG file to extract data from.
	 * @param params  Additional parameters.
	 */
	public ShpFileDune2(String name, PngFile pngfile, String... params) {

		this(name, (ImageFile)pngfile, params);
	}

	/**
	 * Constructor, builds this file from another <code>ImagesFile</code>
	 * implementation.
	 * 
	 * @param name		 Name of this file.
	 * @param imagesfile File to source data from.
	 */
	private ShpFileDune2(String name, ImagesFile imagesfile) {

		this(name, new FilePointer(name, false));

		srcimgsfile = imagesfile;
		destnumimgs = imagesfile.numImages();
		destwidth  = 0;
		destheight = 0;
		for (int i = 0; i < destnumimgs; i++) {
			if (destwidth < srcimgsfile.width(i)) {
				destwidth = srcimgsfile.width(i);
			}
			if (destheight < srcimgsfile.height(i)) {
				destheight = srcimgsfile.height(i);
			}
		}
	}

	/**
	 * Constructor, builds this filetype using the data from the another
	 * <code>ShpFile</code> type.
	 * 
	 * @param name	  The name of this file.
	 * @param shpfile Fully read SHP file to extract data from.
	 */
	public ShpFileDune2(String name, ShpFile shpfile) {

		this(name, (ImagesFile)shpfile);
	}

	/**
	 * @inheritDoc
	 */
	public ByteBuffer[] applyPalette(PaletteFile palette) {

		ByteBuffer[] rgbaimages = new ByteBuffer[numImages()];
		for (int i = 0; i < numImages(); i++) {
			rgbaimages[i] = BufferUtility.newByteBuffer(width(i) * height (i) * format());
		}
		ImageUtility.applyPalette(shpimages, rgbaimages, format(), palette);
		return rgbaimages;
	}

	/**
	 * Ensures that the source files used to save this Dune 2 SHP file are
	 * within it's maximum limits.
	 */
	private void checkSourceFile() {

		// Check number of images
		if (destnumimgs > MAX_NUMIMGS) {
			throw new UnsupportedOperationException(ProgramStrings.getText(Common_Filetypes.IMAGE_MAX_NUM,
					name, ShpFileDune2.class.getName(), Integer.toString(MAX_NUMIMGS)));
		}

		// Check width
		if (destwidth > MAX_WIDTH) {
			throw new UnsupportedOperationException(ProgramStrings.getText(Common_Filetypes.IMAGE_MAX_WIDTH,
					name, ShpFileDune2.class.getName(), Integer.toString(MAX_WIDTH)));
		}

		// Check height
		if (destheight > MAX_HEIGHT) {
			throw new UnsupportedOperationException(ProgramStrings.getText(Common_Filetypes.IMAGE_MAX_HEIGHT,
					name, ShpFileDune2.class.getName(), Integer.toString(MAX_HEIGHT)));
		}

		// Check size of uncompressed image
		int imagesize = destwidth * destheight * (srcimgfile != null ? srcimgfile.format() : srcimgsfile.format());
		if (imagesize > MAX_IMGSIZE) {
			throw new UnsupportedOperationException(ProgramStrings.getText(Common_Filetypes.IMAGE_MAX_SIZE,
					name, ShpFileDune2.class.getName(), Integer.toString(MAX_IMGSIZE)));
		}
	}

	/**
	 * @inheritDoc
	 */
	public int format() {

		return FORMAT_RGBA;
	}

	/**
	 * @inheritDoc
	 */
	public ByteBuffer[] getImages() {

		return shpimages;
	}

	/**
	 * @inheritDoc
	 */
	public int height(int imagenum) {

		return shpimageheaders[imagenum].height & 0xff;
	}

	/**
	 * @inheritDoc
	 */
	public int numImages() {

		return shpfileheader.numimages & 0xffff;
	}

	/**
	 * @inheritDoc
	 */
	public void readFromFile() {

		// Construct file header
		shpfileheader   = new ShpFileHeaderDune2(file);
		shpimageheaders = new ShpImageHeaderDune2[numImages()];
		shpoffsetsize   = shpfileheader.getOffsetType();

		// Construct image data
		shpimages = new ByteBuffer[numImages()];
		for (int i = 0; i < numImages(); i++) {

			// Read the next image header
			file.seek(shpoffsetsize == IMAGE_OFFSET_2 ?
					shpfileheader.imageoffsets[i]:
					shpfileheader.imageoffsets[i] + 2);
			ShpImageHeaderDune2 imageheader = new ShpImageHeaderDune2(file);

			// Compensation for the compression part
			// NOTE: Is this EVER used?!?!?
			if ((imageheader.compression & 1) != 0) {
				file.seek(file.position() + 16);
			}

			ByteBuffer filedata = file.readBytes(imageheader.filesize - 10);
			ByteBuffer imagedata = BufferUtility.newByteBuffer(imageheader.width * imageheader.height);

			// Decide the decompression method required
			if ((~imageheader.compression & 2) != 0) {
				ByteBuffer uncompressed = BufferUtility.newByteBuffer(imageheader.datasize);
				CNCCodec.decodeFormat80(filedata, uncompressed);
				CNCCodec.decodeFormat2(uncompressed, imagedata);
			}
			else {
				CNCCodec.decodeFormat2(filedata, imagedata);
			}

			shpimageheaders[i] = imageheader;
			shpimages[i] = imagedata;
		}
	}

	/**
	 * @inheritDoc
	 */
	public void saveToFile() {

		// Ensure the source file can be converted to a Dune 2 SHP file
		checkSourceFile();

		// Ensure source is indexed/paletted
		ByteBuffer[] srcimages;
		if (srcimgsfile instanceof Paletted) {
			srcimages = srcimgsfile.getImages();
		}
		else if (srcimgfile instanceof Paletted) {
			srcimages = ImageUtility.splitImage(destwidth, destheight, destnumimgs, srcimgfile);
		}
		else if ((srcimgfile instanceof PngFile) && ((PngFile)srcimgfile).isIndexed()) {
			srcimages = ImageUtility.splitImage(destwidth, destheight, destnumimgs,
					srcimgfile.width(), srcimgfile.height(), 1, ((PngFile)srcimgfile).getRawImage());
		}
		else {
			throw new UnsupportedOperationException(ProgramStrings.getText(
					Common_Filetypes.IMAGE_TYPE_PALETTED, srcimgsfile.toString(), getClass().getName()));
		}

		// Start construction of the file
		shpoffsetsize   = IMAGE_OFFSET_4;
		shpimageheaders = new ShpImageHeaderDune2[destnumimgs];
		shpimages       = new ByteBuffer[destnumimgs];

		int offsettotal = 4 + destnumimgs * shpoffsetsize;
		int[] imageoffsets = new int[destnumimgs];

		// Construct image headers for all of the images
		for (int i = 0; i < destnumimgs; i++) {
			int width  = srcimgsfile != null ? srcimgsfile.width(i)  : destwidth;
			int height = srcimgsfile != null ? srcimgsfile.height(i) : destheight;

			// Get the compressed data for this part
			ByteBuffer imagebytes = srcimages[i];
			ByteBuffer compressedbytes = BufferUtility.newByteBuffer((int)(imagebytes.capacity() * 1.5));
			CNCCodec.encodeFormat2(imagebytes, height, compressedbytes);

			// Image header w/ basic Format2 compression
			// TODO: Implement Format80 compression on the image data
			shpimageheaders[i] = new ShpImageHeaderDune2((short)2, (byte)height,
					(short)width, (byte)height, (short)(10 + compressedbytes.limit()), (short)compressedbytes.limit());

			// Compressed image data w/ basic Format2 compression
			shpimages[i] = compressedbytes;

			// Track offset values
			imageoffsets[i] = offsettotal;
			offsettotal += 10 + compressedbytes.limit();
		}

		// Construct an 'offset 4' file header
		shpfileheader = new ShpFileHeaderDune2((short)destnumimgs, imageoffsets, offsettotal);


		// Write file header to disk
		file.writeShort(shpfileheader.numimages);
		for (int i = 0; i < destnumimgs; i++) {
			file.writeInt(shpfileheader.imageoffsets[i]);
		}
		file.writeInt(shpfileheader.endoffile);

		// Write image headers and image data to disk
		for (int i = 0; i < destnumimgs; i++) {

			ShpImageHeaderDune2 imageoffset = shpimageheaders[i];
			file.writeShort(imageoffset.compression);
			file.writeByte(imageoffset.slices);
			file.writeShort(imageoffset.width);
			file.writeByte(imageoffset.height);
			file.writeShort(imageoffset.filesize);
			file.writeShort(imageoffset.datasize);

			file.writeBytes(shpimages[i]);
		}
	}

	/**
	 * Returns some information on this Dune 2 SHP file.
	 * 
	 * @return Dune 2 SHP file info.
	 */
	public String toString() {

		// Check if widths/heights of this file vary
		boolean variedwidth = false;
		int initwidth = width(0);
		for (int i = 1; i < numImages(); i++) {
			if (initwidth != width(i)) {
				variedwidth = true;
				break;
			}
		}
		boolean variedheight = false;
		int initheight = height(0);
		for (int i = 1; i < numImages(); i++) {
			if (initheight != height(i)) {
				variedheight = true;
				break;
			}
		}

		return name + " (Dune 2 SHP file)" +
			"\n  " + ProgramStrings.getText(Common_Filetypes.IMAGE_NUM) + " " + numImages() +
			"\n  " + ProgramStrings.getText(Common_Filetypes.IMAGE_DIMX) + ": " +
					 (variedwidth ? ProgramStrings.getText(Common_Filetypes.IMAGE_DIM_VARIES) : initwidth) +
			"\n  " + ProgramStrings.getText(Common_Filetypes.IMAGE_DIMY) + ": " +
				 	 (variedheight ? ProgramStrings.getText(Common_Filetypes.IMAGE_DIM_VARIES) : initheight) +
			"\n  " + ProgramStrings.getText(Common_Filetypes.IMAGE_COLOURS) + ": 8-" +
				 	 ProgramStrings.getText(Common_Filetypes.IMAGE_COLOURS_BIT);
	}

	/**
	 * @inheritDoc
	 */
	public int width(int imagenum) {

		return shpimageheaders[imagenum].width & 0xffff;
	}
}
