
// ========================================
// Scanner's Java - Red Alert map tile file
// ========================================

package redhorizon.filetypes.tmp;

import redhorizon.filetypes.FileImplementation;
import redhorizon.filetypes.ImagesFile;
import redhorizon.filetypes.Paletted;
import redhorizon.filetypes.PaletteFile;
import redhorizon.utilities.BufferUtility;
import redhorizon.utilities.FilePointer;
import redhorizon.utilities.ImageUtility;

import java.nio.ByteBuffer;

/**
 * Representation of Red Alert's map tiles.  These are the various bits and
 * pieces which comprise the [MapPack] section of the scenario files.
 * 
 * @author Emanuel Rabina
 */
public class TmpFileRA extends FileImplementation implements ImagesFile, Paletted {

	private TmpHeaderRA tmpheader;
	private ByteBuffer[] tmpimages;

	/**
	 * Constructor, uses the given name for the name of the TMP file, then
	 * generates a <code>FilePointer</code> from that name.
	 * 
	 * @param name The name of the WSA file.
	 */
	public TmpFileRA(String name) {

		this(name, new FilePointer(name));
	}

	/**
	 * Constructor, uses the given name for the TMP file, and parses the data
	 * from the <code>FilePointer</code> into a useable form.
	 * 
	 * @param name The name of this file.
	 * @param file <code>FilePointer</code> to the TMP file.
	 */
	public TmpFileRA(String name, FilePointer file) {

		super(name, file);
	}

	/**
	 * @inheritDoc
	 */
	public ByteBuffer[] applyPalette(PaletteFile palette) {

		ByteBuffer[] rgbaimages = new ByteBuffer[numImages()];
		for (int i = 0; i < numImages(); i++) {
			rgbaimages[i] = BufferUtility.newByteBuffer(width(i) * height (i) * format());
		}
		ImageUtility.applyPalette(tmpimages, rgbaimages, format(), palette);
		return rgbaimages;
	}

	/**
	 * @inheritDoc
	 */
	public int format() {

		return FORMAT_RGBA;
	}

	/**
	 * @inheritDoc
	 */
	public ByteBuffer[] getImages() {

		return tmpimages;
	}

	/**
	 * @inheritDoc
	 */
	public int height(int imagenum) {

		return tmpheader.height & 0xffff;
	}

	/**
	 * @inheritDoc
	 */
	public int numImages() {

		return tmpheader.numtiles & 0xffff;
	}

	/**
	 * @inheritDoc
	 */
	public void readFromFile() {

		// Read header
		tmpheader = new TmpHeaderRA(file);

		// Fill images
		tmpimages = new ByteBuffer[numImages()];

		// Read the image offset array
		file.seek(tmpheader.index1);
		byte[] imageoffs = new byte[numImages()];
		for (int i = 0; i < numImages(); i++) {
			imageoffs[i] = file.readByte();
		}

		// Populate the image array with the image data
		for (int i = 0; i < numImages(); i++) {
			byte imageoff = imageoffs[i];

			// Skip empty tiles
			if (imageoff == 0xff) {
				continue;
			}

			// Seek to the position of the image, read
			file.seek(tmpheader.imagedata + (imageoff * height(0) * width(0)));
			tmpimages[i] = file.readBytes(width(0) * height(0));
		}
	}

	/**
	 * @inheritDoc
	 */
	public int width(int imagenum) {

		return tmpheader.width & 0xffff;
	}
}
