
// =============================================
// Scanner's Java - Representation of a VQA file
// =============================================

package redhorizon.filetypes.vqa;

import redhorizon.filetypes.FileImplementation;
import redhorizon.filetypes.Streaming;
import redhorizon.filetypes.VideoFile;
import redhorizon.utilities.BufferUtility;
import redhorizon.utilities.CNCCodec;
import redhorizon.utilities.DataStream;
import redhorizon.utilities.FilePointer;
import redhorizon.utilities.ImageUtility;

import java.nio.ByteBuffer;

/**
 * Implementation of a VQA file.  This is the video format used in Red Alert and
 * Tiberium Dawn.<br>
 * <br>
 * The 2 documents I've obtained conflict slightly with each other as to the
 * interpretation of the file layout.  I'll use my own, combined with the
 * knowledge gained from the old FreeCNC project, to structure the VQA files.
 * 
 * @author Emanuel Rabina
 */
public class VqaFile extends FileImplementation implements VideoFile, Streaming {

	/**
	 * List of known chunks for the VQA file format
	 */
	private enum VqaChunks {

		CHUNK_FORM ("FORM"),
		CHUNK_FINF ("FINF"),
		CHUNK_VQFR ("VQFR"),

		CHUNK_CBF  ("CBF"),
		CHUNK_CBP  ("CBP"),
		CHUNK_CPL  ("CPL"),
		CHUNK_SND  ("SND"),
		CHUNK_VPT  ("VPT");

		private final String chunkname;

		/**
		 * Constructor, matches the enumerated type with the chunk designation.
		 * 
		 * @param chunkname Identifying name of the chunk.
		 */
		private VqaChunks(String chunkname) {

			this.chunkname = chunkname;
		}

		/**
		 * Returns whether the given chunkname is a known chunk.
		 * 
		 * @param name The chunk name to compare against.
		 * @return <code>true</code> if the chunk name is recognized,
		 * 		   <code>false</code> otherwise.
		 */
		private static boolean isKnownChunk(String name) {

			for (VqaChunks vqachunk: VqaChunks.values()) {
				if (name.startsWith(vqachunk.chunkname)) {
					return true;
				}
			}
			return false;
		}
	}

	private static final String CHUNK_UNCOMPRESSED = "0";
	private static final String CHUNK_COMPRESSED   = "Z";
	private static final String TYPE_WS_ADPCM  = "SND1";
	private static final String TYPE_IMA_ADPCM = "SND2";

	private VqaHeader vqaheader;
	private VqaColour[] vqapalette;

	private int blocksize;
	private int blockparts;
	private int framesize;
	private int tablesize;

	/**
	 * Constructor, takes the name of the VQA file and uses that name to
	 * generate a new <code>FilePointer</code>.
	 * 
	 * @param name The name of the VQA file.
	 */
	public VqaFile(String name) {

		this(name, new FilePointer(name));
	}

	/**
	 * Constructor, takes the name of the VQA and then parses the file using the
	 * given <code>FilePointer</code>.
	 * 
	 * @param name The name of this file.
	 * @param file <code>FilePointer</code> to the VQA file.
	 */
	public VqaFile(String name, FilePointer file) {

		super(name, file);
	}

	/**
	 * @inheritDoc
	 */
	public float adjustmentFactor() {

		return 1.2f;
	}

	/**
	 * @inheritDoc
	 */
	public int bitrate() {

		return vqaheader.bits;
	}

	/**
	 * Returns the height of each video block.
	 * 
	 * @return The height of each block.
	 */
	private byte blockHeight() {

		return vqaheader.blockheight;
	}

	/**
	 * Returns the width of each video block.
	 * 
	 * @return The width of each block.
	 */
	private byte blockWidth() {

		return vqaheader.blockwidth;
	}

	/**
	 * @inheritDoc
	 */
	public int channels() {

		return vqaheader.channels;
	}

	/**
	 * Decodes the complete lookup table (CBF chunk) of a VQA file.
	 * 
	 * @param vqachunk Currenlty read <code>VqaChunk</code>.
	 * @param bytes	   <code>ByteBuffer</code> containing the lookup table data.
	 * @return A full lookup table.
	 */
	private VqaTable decodeCBFChunk(VqaChunk vqachunk, ByteBuffer bytes) {

		ByteBuffer tablebytes;

		// Decode uncompressed
		if (vqachunk.chunkname.endsWith(CHUNK_UNCOMPRESSED)) {
			tablebytes = bytes;
		}
		// Decode compressed
		else {
			tablebytes = BufferUtility.newByteBuffer(tablesize);
			CNCCodec.decodeFormat80(bytes, tablebytes);
		}
		return new VqaTable(tablebytes);
	}

	/**
	 * Decodes the partial lookup table (CBP chunk) of a VQA file.
	 * 
	 * @param vqachunk		Currenlty read <code>VqaChunk</code>.
	 * @param bytes			<code>ByteBuffer</code> containing the partial
	 * 						lookup data.
	 * @param partialtables Partially constructed lookup table.
	 * @return <code>true</code> if the partials are compressed,
	 * 		   <code>false</code> otherwise.
	 */
	private boolean decodeCBPChunk(VqaChunk vqachunk, ByteBuffer bytes, ByteBuffer[] partialtables) {

		// Append the new partial to the array of partials
		for (int i = 0; i < partialtables.length; i++) {
			if (partialtables[i] == null) {
				partialtables[i] = bytes;
				break;
			}
		}

		// Return whether compressed or not
		return vqachunk.chunkname.endsWith(CHUNK_COMPRESSED);
	}

	/**
	 * Decodes the palette data (CPL chunk) of a VQA file.
	 * 
	 * @param vqachunk Currenlty read <code>VqaChunk</code>.
	 * @param bytes	   <code>ByteBuffer</code> containing the palette data.
	 * @return The palette from this chunk.
	 */
	private VqaColour[] decodeCPLChunk(VqaChunk vqachunk, ByteBuffer bytes) {

		ByteBuffer palbytes;

		// Decode uncompressed
		if (vqachunk.chunkname.endsWith(CHUNK_UNCOMPRESSED)) {
			palbytes = bytes;
		}
		// Decode compressed
		else {
			palbytes = BufferUtility.newByteBuffer(blocksize);
			CNCCodec.decodeFormat80(bytes, palbytes);
		}

		// Add the colours to the palette
		VqaColour[] newpalette = new VqaColour[numColours()];
		for (int i = 0; i < numColours(); i++) {
			newpalette[i] = new VqaColour(palbytes);
		}
		return newpalette;
	}

	/**
	 * Decodes the SND chunk of a VQA file.
	 * 
	 * @param vqachunk Currenlty read <code>VqaChunk</code>.
	 * @param bytes	   <code>ByteBuffer</code> containing the sound data.
	 * @param update   2-<code>int</code> array of the latest index and sample
	 * 				   values respectively.
	 * @return The sound data from this chunk.
	 */
	private ByteBuffer decodeSNDChunk(VqaChunk vqachunk, ByteBuffer bytes, int[] update) {

		ByteBuffer soundbytes;

		// Decode uncompressed
		if (vqachunk.chunkname.endsWith(CHUNK_UNCOMPRESSED)) {
			soundbytes = bytes;
		}
		// Decode compressed
		else {
			soundbytes = BufferUtility.newByteBuffer(vqachunk.length << 2);

			if (vqachunk.chunkname.equals(TYPE_WS_ADPCM)) {
				CNCCodec.decompress8bitWSADPCM(bytes, soundbytes);
			}
			else if (vqachunk.chunkname.equals(TYPE_IMA_ADPCM)) {
				CNCCodec.decompress16bitIMAADPCM(bytes, soundbytes, update);
			}
		}
		return soundbytes;
	}

	/**
	 * Decodes the video data (VPT chunk) of a VQA file.
	 * 
	 * @param vqachunk	 Currenlty read <code>VqaChunk</code>.
	 * @param bytes		 <code>ByteBuffer</code> containing the sound data.
	 * @param vqatable	 Current values of the lookup table.
	 * @return A fully decoded frame of video.
	 */
	private ByteBuffer decodeVPTChunk(VqaChunk vqachunk, ByteBuffer bytes, VqaTable vqatable) {

		ByteBuffer videobytes;

		// Decode uncompressed
		if (vqachunk.chunkname.endsWith(CHUNK_UNCOMPRESSED)) {
			videobytes = bytes;
		}
		// Decode compressed
		else {
			videobytes = BufferUtility.newByteBuffer(blockparts << 1);
			CNCCodec.decodeFormat80(bytes, videobytes);
		}

		ByteBuffer framebytes = BufferUtility.newByteBuffer(framesize);

		// Now decode every block
		int nextline = width() - blockWidth();
		int modifier = blockHeight() == 2 ? 0xf : 0xff;
		int block = 0;
		int offset = blockparts;

		// Go across first, then down
		for (int y = 0; y < height(); y += blockHeight()) {
			for (int x = 0; x < width(); x += blockWidth(), block++) {

				framebytes.position(y * width() + x);

				// Get the proper lookup value for the block
				int topval = videobytes.get(block) & 0xff;
				int botval = videobytes.get(block + offset) & 0xff;

				// Fill the block with 1 colour
				if (botval == modifier) {

					for (int i = 1; i <= blocksize; i++) {
						framebytes.put((byte)topval);

						if ((i % blockWidth() == 0) && (i != blocksize)) {
							framebytes.position(framebytes.position() + nextline);
						}
					}
				}

				// Otherwise, fill the block with the one in the lookup table
				else {
					int ref = ((botval << 8) + topval) * blocksize;

					for (int i = 1; i <= blocksize; i++) {
						framebytes.put(vqatable.getValueAt(ref++));

						if ((i % blockWidth() == 0) && (i != blocksize)) {
							framebytes.position(framebytes.position() + nextline);
						}
					}
				}
			}
		}
		framebytes.rewind();

		// Apply the palette
		ByteBuffer colourbytes = BufferUtility.newByteBuffer(framesize * 3);
		ImageUtility.applyPalette(framebytes, colourbytes, vqapalette);

		return colourbytes;
	}

	/**
	 * @inheritDoc
	 */
	public int format() {

		return 3;
	}

	/**
	 * Returns the frames/second at which this VQA file plays.
	 * 
	 * @return Frames/second of the video.
	 */
	public float frameRate() {

		return vqaheader.framerate;
	}

	/**
	 * @inheritDoc
	 */
	public int frequency() {

		return vqaheader.frequency;
	}

	/**
	 * @inheritDoc
	 */
	public ByteBuffer[] getImages() {

		return null;
	}

	/**
	 * @inheritDoc
	 */
	public DataStream getPipe() {

		// Creates the decoder algorithm and applies it to the pipe
		final DataStream pipe = new DataStream(2);
		pipe.setAlgorithm(new Runnable() {
			public void run() {

				// Video decoding components
				VqaChunk vqachunk;
				VqaTable vqatable = null;
				ByteBuffer[] partialtables = new ByteBuffer[numLookupTables()];
				boolean compressed = false;

				ByteBuffer vqasound = null;
				ByteBuffer vqaframe = null;
				int[] update = {0,0};

				// Video decoding loop
				for (int i = -1, j = 0; j < numImages(); ) {

					// Use the generic chunk to determine the next path
					vqachunk = readNextChunk();

					// Decode sound data
					if (vqachunk.chunkname.startsWith(VqaChunks.CHUNK_SND.chunkname)) {
						vqasound = decodeSNDChunk(vqachunk, file.readBytes(vqachunk.length), update);
						i++;
					}

					// Decode video data
					else if (vqachunk.chunkname.equals(VqaChunks.CHUNK_VQFR.chunkname)) {

						while (true) {
							vqachunk = readNextChunk();

							ByteBuffer videobytes = file.readBytes(vqachunk.length);

							// Full lookup table
							if (vqachunk.chunkname.startsWith(VqaChunks.CHUNK_CBF.chunkname)) {
								vqatable = decodeCBFChunk(vqachunk, videobytes);
							}
							// Partial lookup table
							else if (vqachunk.chunkname.startsWith(VqaChunks.CHUNK_CBP.chunkname)) {
								compressed = decodeCBPChunk(vqachunk, videobytes, partialtables);
							}
							// Palette
							else if (vqachunk.chunkname.startsWith(VqaChunks.CHUNK_CPL.chunkname)) {
								vqapalette = decodeCPLChunk(vqachunk, videobytes);
							}
							// Video data
							else if (vqachunk.chunkname.startsWith(VqaChunks.CHUNK_VPT.chunkname)) {
								vqaframe = decodeVPTChunk(vqachunk, videobytes, vqatable);
								break;
							}
						}

						// If full, replace the old lookup table
						if (partialtables[numLookupTables() - 1] != null) {
							vqatable = replaceLookupTable(partialtables, compressed);
						}
						j++;
					}

					// Write to the input stream when there's data for both frame and sound
					if (i == j) {
						pipe.getSink().write(new ByteBuffer[]{ vqaframe, vqasound });
					}
				}
				pipe.getSink().finish();
			}
		});
		return pipe;
	}

	/**
	 * @inheritDoc
	 */
	public ByteBuffer getSound() {

		return null;
	}

	/**
	 * @inheritDoc
	 */
	public int height() {

		return vqaheader.height & 0xffff;
	}

	/**
	 * Returns the maximum number of blocks used in the compression of this
	 * video.
	 * 
	 * @return The maximum number of blocks per frame.
	 */
	private int maxBlocks() {

		return vqaheader.maxblocks;
	}

	/**
	 * Returns the number of colours used for this video.
	 * 
	 * @return Number of colours in this video.
	 */
	private int numColours() {

		return vqaheader.numcolours & 0xffff;
	}

	/**
	 * @inheritDoc
	 */
	public int numImages() {

		return vqaheader.numframes & 0xffff;
	}

	/**
	 * Returns the number of partial lookup tables required to make a complete
	 * one.
	 * 
	 * @return The number of frames which share a lookup table.
	 */
	private int numLookupTables() {

		return vqaheader.cbparts & 0xff;
	}

	/**
	 * @inheritDoc
	 */
	public void readFromFile() {

		// Write header
		readNextChunk();
		vqaheader  = new VqaHeader(file);

		// Write offsets
		readNextChunk();
		for (int i = 0; i < numImages(); i++) {
			vqaheader.addOffset(new VqaOffset(file));
		}

		// Set some useful constants
		blocksize  = blockWidth() * blockHeight();
		blockparts = (width() / blockWidth()) * (height() / blockHeight());
		framesize  = width() * height();
		tablesize  = maxBlocks() * blocksize;
	}

	/**
	 * Reads the next chunk of data, bypassing the null bytes used to make the
	 * offsets all even.
	 * 
	 * @return The next <code>VqaChunk</code>.
	 */
	private VqaChunk readNextChunk() {

		VqaChunk vqachunk;
		while (true) {

			// Skip null byte
			if (file.readByte() != 0) {
				file.seek(file.position() - 1);
			}
			vqachunk = new VqaChunk(file);

			// Break only on known chunk names
			if (VqaChunks.isKnownChunk(vqachunk.chunkname)) {
				break;
			}

			file.seek(file.position() + vqachunk.length);
		}
		return vqachunk;
	}

	/**
	 * Replaces the VQA lookup table with the complete construction of several
	 * partials.
	 * 
	 * @param partials	 Array of partials so far.
	 * @param compressed Whether or not the array of partials is compressed.
	 * @return The replacement lookup table.
	 */
	private VqaTable replaceLookupTable(ByteBuffer[] partials, boolean compressed) {

		// Generate complete table
		ByteBuffer fulltable = BufferUtility.newByteBuffer(tablesize);
		for (int i = 0; i < numLookupTables(); i++) {
			fulltable.put(partials[i]);
			partials[i] = null;
		}
		fulltable.rewind();

		// Decode uncompressed
		ByteBuffer tablebytes;
		if (compressed) {
			tablebytes = BufferUtility.newByteBuffer(tablesize);
			CNCCodec.decodeFormat80(fulltable, tablebytes);
		}
		// Decode compressed
		else {
			tablebytes = fulltable;
		}
		return new VqaTable(tablebytes);
	}

	/**
	 * @inheritDoc
	 */
	public int width() {

		return vqaheader.width & 0xffff;
	}
}
