
// =============================================
// Scanner's Java - Representation of a WSA file
// =============================================

package redhorizon.filetypes.wsa;

import redhorizon.filetypes.AnimationFile;
import redhorizon.filetypes.FileImplementation;
import redhorizon.filetypes.Streaming;
import redhorizon.utilities.BufferUtility;
import redhorizon.utilities.CNCCodec;
import redhorizon.utilities.DataStream;
import redhorizon.utilities.FilePointer;
import redhorizon.utilities.ImageUtility;

import java.nio.ByteBuffer;

/**
 * Implementation of a WSA file, which is the format of short animations in Red
 * Alert and Tiberium Dawn.  WSA files are generally large, so implements the
 * {@link Streaming} interface for streaming decoding and playback.
 * 
 * @author Emanuel Rabina
 */
public class WsaFile extends FileImplementation implements AnimationFile, Streaming {

	private WsaHeader wsaheader;
	private WsaColour[] wsapalette;

	/**
	 * Constructor, uses the given name for the name of the WSA file, then
	 * generates a <code>FilePointer</code> from that name.
	 * 
	 * @param name The name of the WSA file.
	 */
	public WsaFile(String name) {

		this(name, new FilePointer(name));
	}

	/**
	 * Constructor, uses the given name for the WSA file, and parses the data
	 * from the <code>FilePointer</code> into a useable form.
	 * 
	 * @param name The name of this file.
	 * @param file <code>FilePointer</code> to the WSA file.
	 */
	public WsaFile(String name, FilePointer file) {

		super(name, file);
	}

	/**
	 * @inheritDoc
	 */
	public float adjustmentFactor() {

		return 1.2f;
	}

	/**
	 * Returns the coordinates for the upper-left corner of the animation.  Used
	 * particularly on widescreen animations where they don't cover the whole
	 * screen.
	 * 
	 * @return 2-element <code>int</code>[] of the x & y coordinate for the
	 * 		   upper-left corner.
	 */
	public int[] cornerCoord() {

		return new int[]{ wsaheader.x, wsaheader.y };
	}

	/**
	 * Returns the offset information for the given frame number.
	 * 
	 * @param frame The frame number we're after.
	 * @return The matching offset information.
	 */
	private WsaOffset imageOffset(int frame) {

		return wsaheader.offsets[frame];
	}

	/**
	 * @inheritDoc
	 */
	public int format() {

		return 3;
	}

	/**
	 * @inheritDoc
	 */
	public float frameRate() {

		// Below is the documented method for obtaining the framerate for WSA
		// files, but the returned result feels too slow
//		return 1f / (wsaheader.delta / 1024f) * 1000f;
		return 20f;
	}

	/**
	 * @inheritDoc
	 */
	public ByteBuffer[] getImages() {

		return null;
	}

	/**
	 * @inheritDoc
	 */
	public DataStream getPipe() {

		// Prepare a new streaming decoding thread
		final DataStream pipe = new DataStream(1);
		pipe.setAlgorithm(new Runnable() {
			public void run() {

				Thread.currentThread().setName("WSA file decoding thread - " + name);
				ByteBuffer lastbytes = null;

				// Decompress every frame
				for (int i = 0; i < numImages(); i++) {

					// Information on the current frame
					WsaOffset frameoffset = imageOffset(i);
					int sourceoffset = frameoffset.offset;

					// Format conversion buffers
					int sourcelength = imageOffset(i + 1).offset - sourceoffset;
					ByteBuffer sourcebytes = file.readBytes(sourcelength);

					int destlength = width() * height();
					ByteBuffer intbytes   = BufferUtility.newByteBuffer(destlength);
					ByteBuffer framebytes = BufferUtility.newByteBuffer(destlength);

					// First decompress from Format80...
					CNCCodec.decodeFormat80(sourcebytes, intbytes);

					// ...then decode as Format40
					if (i != 0) {
						framebytes.put(lastbytes).rewind();
					}
					CNCCodec.decodeFormat40(intbytes, framebytes);

					// Apply palette to the frame
					ByteBuffer colourbytes = BufferUtility.newByteBuffer(destlength * 3);
					ImageUtility.applyPalette(framebytes, colourbytes, wsapalette);

					// Remember the last frame for Decode40 'difference' encoding
					lastbytes = framebytes;
					lastbytes.rewind();

					// Send the decompressed image down the data stream
					pipe.getSink().write(new ByteBuffer[]{ colourbytes });
				}
				pipe.getSink().finish();
			}
		});
		return pipe;
	}

	/**
	 * @inheritDoc
	 */
	public int height() {

		return wsaheader.height & 0xffff;
	}

	/**
	 * @inheritDoc
	 */
	public int numImages() {

		return wsaheader.numframes & 0xffff;
	}

	/**
	 * @inheritDoc
	 */
	public void readFromFile() {

		wsaheader = new WsaHeader(file);

		// numImages() + 2 for the 0 offset and EOF pointer
		for (int i = 0; i < numImages() + 2; i++) {
			WsaOffset wsaoffset = new WsaOffset(file);
			wsaheader.addOffSet(wsaoffset);
		}

		// Add internal palette
		wsapalette = new WsaColour[256];
		for (int i = 0; i < 256; i++) {
			wsapalette[i] = new WsaColour(file);
		}
	}

	/**
	 * @inheritDoc
	 */
	public int width() {

		return wsaheader.width & 0xffff;
	}
}
