
// ===========================================
// Scanner's Java - Cube used to depict volume
// ===========================================

package redhorizon.misc.geometry;

/**
 * Immutable representation of a cube, used to depict 3-dimensional volume.  The
 * XYZ co-ordinates represent the center of the cube, rather than an arbitrary
 * corner.
 * 
 * @author Emanuel Rabina
 */
public class Cube6i extends Rectangle4i implements Cloneable {

	final int z;
	final int depth;

	/**
	 * Constructor, creates a cube of the given width, height, and depth, but
	 * with no current XYZ co-ordinates.  Used for cubes of known volume, but no
	 * position in space.
	 * 
	 * @param width	 Width of the cube.
	 * @param height Height of the cube.
	 * @param depth	 Depth of the cube.
	 */
	public Cube6i(int width, int height, int depth) {

		this(0, 0, 0, width, height, depth);
	}

	/**
	 * Constructor, creates a cube with both volume and positional information.
	 * 
	 * @param coords Initial XYZ co-ordinates.
	 * @param width	 Width of the cube.
	 * @param height Height of the cube.
	 * @param depth	 Depth of the cube.
	 */
	public Cube6i(Point3i coords, int width, int height, int depth) {

		this(coords.x, coords.y, coords.z, width, height, depth);
	}

	/**
	 * Constructor, creates a cube with both volume and positional information.
	 * 
	 * @param x		 X co-ordinate.
	 * @param y		 Y co-ordinate.
	 * @param z		 Z co-ordinate.
	 * @param width	 Width of the cube.
	 * @param height Height of the cube.
	 * @param depth	 Depth of the cube.
	 */
	public Cube6i(int x, int y, int z, int width, int height, int depth) {

		super(x, y, width, height);
		this.z     = z;
		this.depth = depth;
	}

	/**
	 * Returns a cube whose volume is the total width height and depth required
	 * to encompass both this and the given cube.
	 * 
	 * @param append Volume to append to this one.
	 */
	public Cube6i append(Cube6i append) {

		int newx = (x + append.x) >> 1;
		int newy = (y + append.y) >> 1;
		int newz = (z + append.z) >> 1;
		int totalw = (width >> 1)  + (append.width >> 1)  + (Math.abs(x - append.x));
		int totalh = (height >> 1) + (append.height >> 1) + (Math.abs(y - append.y));
		int totalz = (depth >> 1)  + (append.depth >> 1)  + (Math.abs(z - append.z));

		return new Cube6i(newx, newy, newz, totalw, totalh, totalz);
	}

	/**
	 * Returns a copy of this <code>Cube6i</code>.
	 * 
	 * @return Copy of this instance.
	 */
	public Object clone() {

		return new Cube6i(x, y, z, width, height, depth);
	}

	/**
	 * Returns the depth of this cube.
	 * 
	 * @return Depth of this cube.
	 */
	public int getDepth() {

		return depth;
	}

	/**
	 * Returns the Z co-ordinate of this cube.
	 * 
	 * @return Z co-ordinate.
	 */
	public int getZ() {

		return z;
	}

	/**
	 * Returns a new cube whose XYZ co-ordinates are of this cube shifted by the
	 * specified amount.
	 * 
	 * @param shift XYZ co-ordinates describing the difference between the
	 * 				current position and the desired position.
	 */
	public Cube6i shift(Point3i shift) {

		return new Cube6i(x + shift.x, y + shift.y, z + shift.z, width, height, depth);
	}

	/**
	 * Returns this cube represented as ((x,y,z),(width,height,depth)).
	 * 
	 * @return Cubei((x,y,z),(width,height,depth)).
	 */
	public String toString() {

		return "Cube6i((" + x + "," + y + "," + z + "),(" + width + "," + height + "," + depth + "))";
	}
}
