
// =========================
// Scanner's Java - 2D point
// =========================

package redhorizon.misc.geometry;

/**
 * Immutable representation of a point in 2-dimensional space.
 * 
 * @author Emanuel Rabina
 */
public class Point2i implements Cloneable {

	// Default no co-ordinates point, used mainly for comparisons
	public static final Point2i DEFAULT = new Point2i(0,0);

	final int x;
	final int y;

	/**
	 * Constructor, assigns the values of the X & Y co-ordinates.
	 * 
	 * @param x X co-ordinate.
	 * @param y Y co-ordinate.
	 */
	public Point2i(int x, int y) {

		this.x = x;
		this.y = y;
	}

	/**
	 * Returns a new <code>Point2i</code> whose co-ordinates are the result of
	 * adding this point's co-ordinates to those of the <code>Point2i</code>
	 * specified.
	 * 
	 * @param add The other point object to add by.
	 * @return New point from the result of the addition.
	 */
	public Point2i add(Point2i add) {

		return add(add.x, add.y);
	}

	/**
	 * Returns a new <code>Point2i</code> whose co-ordinates are the result of
	 * adding this point's co-ordinates to the x and y values specified.
	 * 
	 * @param addx Amount to add to the X value.
	 * @param addy Amount to add to the Y value.
	 * @return New point from the result of the addition.
	 */
	public Point2i add(int addx, int addy) {

		return new Point2i(x + addx, y + addy);
	}

	/**
	 * Returns a copy of this <code>Point2i</code> instance.
	 * 
	 * @return Copy of this instance.
	 */
	public Object clone() {

		return new Point2i(x, y);
	}

	/**
	 * Returns a new <code>Point2i</code> with XY values set to the difference
	 * between this point and the given point.
	 * 
	 * @param point The other point to compare distances against.
	 * @return New <code>Point2i</code> of the distance between the 2 points.
	 */
	public Point2i difference(Point2i point) {

		return difference(point.x, point.y);
	}

	/**
	 * Returns a new <code>Point2i</code> with XY values set to the difference
	 * between this point and the given XY values.
	 * 
	 * @param diffx An X value to compare against.
	 * @param diffy A Y value to compare against.
	 * @return New <code>Point2i</code> of the distance between the 2 points.
	 */
	public Point2i difference(int diffx, int diffy) {

		return new Point2i(x - diffx, y - diffy);
	}

	/**
	 * A <code>Point2i</code> is considered equivalent to another if they share
	 * the same XY co-ordinates.
	 * 
	 * @param obj The other <code>Point2i</code> with which to compare.
	 * @return <code>true</code> if the 2 points share the same co-ordinates,
	 * 		   <code>false</code> otherwise.
	 */
	public boolean equals(Object obj) {

		if (obj == this) {
			return true;
		}

		if (obj instanceof Point2i) {
			Point2i other = (Point2i)obj;
			return x == other.x && y == other.y;
		}

		return false;
	}

	/**
	 * Returns the X co-ordinate of this point.
	 * 
	 * @return X co-ordinate.
	 */
	public int getX() {

		return x;
	}

	/**
	 * Returns the Y co-ordinate of this point.
	 * 
	 * @return Y co-ordinate.
	 */
	public int getY() {

		return y;
	}

	/**
	 * Re-defines the <code>hashCode</code> function performed on this object.
	 * 
	 * @return A hashcode value for the <code>Point2i</code>, such that equal
	 * 		   co-ordinates will have the same hash value.
	 */
	public int hashCode() {

		return (y * Short.MAX_VALUE) + x;
	}

	/**
	 * Returns the XY values as an array.  Used primarily for OpenGL/AL array
	 * functions.
	 * 
	 * @return XY values, as the 0th and 1st values in the returned 3-int array.
	 */
	public int[] toArray() {

		return new int[]{ x, y };
	}

	/**
	 * Returns this point represented as (x,y).
	 * 
	 * @return Point2i(x,y).
	 */
	public String toString() {

		return "Point2i(" + x + "," + y + ")";
	}
}
