
// ======================================================
// Scanner's Java - Representation of a point in 3D space
// ======================================================

package redhorizon.misc.geometry;

/**
 * Immutable representation of a point in 3-dimensional space.
 * 
 * @author Emanuel Rabina
 */
public class Point3i extends Point2i {

	// Default no co-ordinates point, used mainly for comparisons
	@SuppressWarnings("hiding")
	public static final Point3i DEFAULT = new Point3i(0,0,0);

	final int z;

	/**
	 * Constructor, assigns the values of all of the co-ordinate variables.
	 * 
	 * @param x X co-ordinate.
	 * @param y Y co-ordinate.
	 * @param z Z co-ordinate.
	 */
	public Point3i(int x, int y, int z) {

		super(x, y);
		this.z = z;
	}

	/**
	 * Returns a new <code>Point3i</code> whose co-ordinates are the result of
	 * adding this point's co-ordinates to those of the <code>Point3i</code>
	 * specified.
	 * 
	 * @param add The other point object to add by.
	 * @return New point from the result of the addition.
	 */
	public Point3i add(Point3i add) {

		return add(add.x, add.y, add.z);
	}

	/**
	 * Returns a new <code>Point3i</code> whose co-ordinates are the result of
	 * adding this point's co-ordinates to the X, Y and Z values specified.
	 * 
	 * @param addx Amount to add to the X value.
	 * @param addy Amount to add to the Y value.
	 * @param addz Amount to add to the Z value.
	 * @return New point from the result of the addition.
	 */
	public Point3i add(int addx, int addy, int addz) {

		return new Point3i(x + addx, y + addy, z + addz);
	}

	/**
	 * Returns a copy of this <code>Point3i</code> instance.
	 * 
	 * @return Copy of this instance.
	 */
	public Object clone() {

		return new Point3i(x, y, z);
	}

	/**
	 * Returns a new <code>Point3i</code> with XYZ values set to the difference
	 * between this point and the given point.
	 * 
	 * @param point The other point to compare distances against.
	 * @return New <code>Point3i</code> of the distance between the 2 points.
	 */
	public Point3i difference(Point3i point) {

		return difference(point.x, point.y, point.z);
	}

	/**
	 * Returns a new <code>Point3i</code> with XYZ values set to the difference
	 * between this point and the given point.
	 * 
	 * @param diffx An X value to compare against.
	 * @param diffy A Y value to compare against.
	 * @param diffz A Z value to compare against.
	 * @return New <code>Point3i</code> of the distance between the 2 points.
	 */
	public Point3i difference(int diffx, int diffy, int diffz) {

		return new Point3i(x - diffx, y - diffy, z - diffz);
	}

	/**
	 * A <code>Point3i</code> is considered equivalent to another if they share
	 * the same XYZ co-ordinates.
	 * 
	 * @param obj The other <code>Point3i</code> with which to compare.
	 * @return <code>true</code> if the 2 points share the same co-ordinates,
	 * 		   <code>false</code> otherwise.
	 */
	public boolean equals(Object obj) {

		if (obj == this) {
			return true;
		}

		if (obj instanceof Point3i) {
			Point3i other = (Point3i)obj;
			return x == other.x && y == other.y && z == other.z;
		}

		return false;
	}

	/**
	 * Returns the Z co-ordinate of this point.
	 * 
	 * @return Z co-ordinate.
	 */
	public int getZ() {

		return z;
	}

	/**
	 * Re-defines the <code>hashCode</code> function performed on this object.
	 * Note that a <code>Point3i</code> with the same co-ordinates as a
	 * <code>Point2i</code> will return the same hash code.
	 * 
	 * @return A hashcode value for the <code>Point3i</code>, such that equal
	 * 		   co-ordinates will have the same hash value.
	 */
	public int hashCode() {

		return (z * Short.MAX_VALUE ^ 2) + (y * Short.MAX_VALUE) + x;
	}

	/**
	 * Returns the XYZ values as an array.  Used primarily for OpenGL/AL array
	 * functions.
	 * 
	 * @return XYZ values, as the 0th, 1st, and 2nd values in the returned 3-int
	 * 		   array.
	 */
	public int[] toArray() {

		return new int[]{ x, y, z };
	}

	/**
	 * Returns this point represented as (x,y,z).
	 * 
	 * @return Point3i(x,y,z).
	 */
	public String toString() {

		return "Point3i(" + x + "," + y + "," + z + ")";
	}
}
