
// =====================================================
// Scanner's Java - Rectangle area used to depict bounds
// =====================================================

package redhorizon.misc.geometry;

/**
 * Immutable representation of a rectangle, used to depict an area or bounds.
 * The XY co-ordinates represent the center of the rectangle, rather than some
 * arbitrary corner.
 * 
 * @author Emanuel Rabina
 */
public class Rectangle4i implements Cloneable {

	final int x;
	final int y;
	final int width;
	final int height;

	/**
	 * Constructor, creates a rectangle of the given width and height, but with
	 * no current XY co-ordinates.  Used for rectangles of known area, but no
	 * position in space.
	 * 
	 * @param width	 Width of the rectangle.
	 * @param height Height of the rectangle.
	 */
	public Rectangle4i(int width, int height) {

		this(0, 0, width, height);
	}

	/**
	 * Constructor, creates a rectangle with both space and area information.
	 * 
	 * @param coords XY co-ordinates.
	 * @param width	 Width of the rectangle.
	 * @param height Height of the rectangle.
	 */
	public Rectangle4i(Point2i coords, int width, int height) {

		this(coords.x, coords.y, width, height);
	}

	/**
	 * Constructor, creates a rectangle with both space and area information.
	 * 
	 * @param x		 X co-ordinate.
	 * @param y		 Y co-ordinate.
	 * @param width	 Width of the rectangle.
	 * @param height Height of the rectangle.
	 */
	public Rectangle4i(int x, int y, int width, int height) {

		this.x      = x;
		this.y      = y;
		this.width  = width;
		this.height = height;
	}

	/**
	 * Returns a rectangle whose area is the total width and height required to
	 * encompass both this and the given rectangle.
	 * 
	 * @param append Area to append to this one.
	 */
	public Rectangle4i append(Rectangle4i append) {

		int newx = (x + append.x) >> 1;
		int newy = (y + append.y) >> 1;
		int totalw = (width >> 1)  + (append.width >> 1)  + (Math.abs(x - append.x));
		int totalh = (height >> 1) + (append.height >> 1) + (Math.abs(y - append.y));

		return new Rectangle4i(newx, newy, totalw, totalh);
	}

	/**
	 * Returns a copy of this <code>Rectangle4i</code>.
	 * 
	 * @return Copy of this instance.
	 */
	public Object clone() {

		return new Rectangle4i(x, y, width, height);
	}

	/**
	 * Creates a rectangle from the given <code>String</code>.
	 * 
	 * @param specs Specs of the intended rectangle, in string format.
	 * @return A rectangle of the given specs.
	 * @throws IllegalArgumentException If the specs string contains an invalid
	 * 		   number of rectangle parts.
	 */
	public static Rectangle4i createFromString(String specs) {

		String specparts = specs.substring(specs.indexOf('(') + 1, specs.indexOf(')'));
		String[] args = specparts.split(",");

		// Delegate to the appropriate constructor
		switch (args.length) {
		case 2:
			return new Rectangle4i(Integer.parseInt(args[0]), Integer.parseInt(args[1]));
		case 4:
			return new Rectangle4i(Integer.parseInt(args[0]), Integer.parseInt(args[1]),
								   Integer.parseInt(args[2]), Integer.parseInt(args[3]));
		default:
			throw new IllegalArgumentException();
		}
	}

	/**
	 * Returns the height of this rectangle.
	 * 
	 * @return Rectangle height.
	 */
	public int getHeight() {

		return height;
	}

	/**
	 * Returns the width of this rectangle.
	 * 
	 * @return Rectangle width.
	 */
	public int getWidth() {

		return width;
	}

	/**
	 * Returns the X co-ordinate of this rectangle.
	 * 
	 * @return X co-ordinate.
	 */
	public int getX() {

		return x;
	}

	/**
	 * Returns the Y co-ordinate of this rectangle.
	 * 
	 * @return Y co-ordinate.
	 */
	public int getY() {

		return y;
	}

	/**
	 * Determines whether the X & Y values of a given point are encompassed by
	 * the 2D area represented by this rectangle.  If the co-ordinates are right
	 * on the boundary of the rectangle, it is not included.
	 * 
	 * @param coord The co-ordinate to test.
	 * @return <code>true</code> if the point is within the rectangle area,
	 * 		   <code>false</code> otherwise.
	 */
	public boolean includes(Point2i coord) {

		return includes(coord.getX(), coord.getY());
	}

	/**
	 * Determines whether the given X & Y values of are encompassed by the 2D
	 * area represented by this rectangle.  If the co-ordinates are right on the
	 * boundary of the rectangle, it is not included.
	 * 
	 * @param coordx X co-ordinate to test.
	 * @param coordy Y co-ordinate to test.
	 * @return <code>true</code> if the point is within the rectangle area,
	 * 		   <code>false</code> otherwise.
	 */
	public boolean includes(int coordx, int coordy) {

		int limitx = x - (width >> 1);
		int limity = y - (height >> 1);
		int limitw = x + (width >> 1);
		int limith = y + (height >> 1);

		return (coordx > limitx && coordx < limitw) &&
			   (coordy > limity && coordy < limith);
	}

	/**
	 * Determines whether the 2D area of another rectangle intersects the 2D
	 * area of this rectangle.
	 * 
	 * @param area The area to test.
	 * @return <code>true</code> if some part intersects this 2D area,
	 * 		   <code>false</code> otherwise.
	 */
	public boolean intersects(Rectangle4i area) {

		int limitx = area.x - (area.width >> 1);
		int limity = area.y - (area.height >> 1);
		int limitw = area.x + (area.width >> 1);
		int limith = area.y + (area.height >> 1);

		Point2i cortl = new Point2i(limitx, limith);
		Point2i corbl = new Point2i(limitx, limity);
		Point2i cortr = new Point2i(limitw, limith);
		Point2i corbr = new Point2i(limitw, limity);

		return includes(cortl) || includes(corbl) || includes(cortr) || includes(corbr);
	}

	/**
	 * Returns a new rectangle whose XY co-ordinates are of this rectangle
	 * shifted by the specified amount.
	 * 
	 * @param shift XY co-ordinates describing the difference between the
	 * 				current position and the desired position.
	 */
	public Rectangle4i shift(Point2i shift) {

		return new Rectangle4i(x + shift.x, y + shift.y, width, height);
	}

	/**
	 * Returns this rectangle represented as ((x,y),(width,height)).
	 * 
	 * @return Rectangle4i((x,y),(width,height)).
	 */
	public String toString() {

		return "Rectangle4i((" + x + "," + y + "),(" + width + "," + height + "))";
	}
}
