
// =====================================
// Scanner's Java - Global game settings
// =====================================

package redhorizon.settings;

import redhorizon.filetypes.SettingsFile;
import redhorizon.filetypes.ini.IniFile;
import redhorizon.utilities.FileUtility;

import java.nio.CharBuffer;
import java.util.HashMap;

/**
 * This class contains all the settings for the startup of the game.  This does
 * not include the settings normally found in Red Alert's Rules.ini file.
 * Instead, it is startup information and any custom information created via the
 * use of Red Alert mods or Red Horizon mods.<br>
 * <br>
 * It acts somewhat like the {@link IniFile} type, except there are
 * additional methods associated with the nature of a global settings file.
 * 
 * @author Emanuel Rabina
 */
public class Settings {

	private static SettingsFile settings;
	private static HashMap<String, HashMap<String, String>> newsettings =
		new HashMap<String, HashMap<String, String>>();

	/**
	 * Hidden default constructor, as this class is only ever meant to be used
	 * statically.
	 */
	private Settings() {
	}

	/**
	 * Saves the settings to the settings file, and then closes links to the
	 * file which stored the settings.
	 */
	public static void close() {

		saveSettings();
		settings = null;
	}

	/**
	 * Returns the value of a given parameter.  The section name is derived from
	 * the type of the <code>Enum</code> used for the parameter, whereas the
	 * parameter name is taken from the <code>toString()</code> method for that
	 * enum.<br>
	 * Enums used to pull out a value must be named
	 * <code>(something)Settings</code>, where the 'something' part is the same
	 * as the section header in the settings file.
	 * 
	 * @param settingkey The parameter the value is mapped to.
	 * @return The value of the setting at the given section and parameter, or
	 * 		   <code>null</code> if no such parameter exists.
	 */
	public static String getSetting(Enum<?> settingkey) {

		String section = settingkey.getClass().getSimpleName();
		section = "[" + section.substring(0, section.indexOf("Settings")) + "]";
		String parameter = settingkey.toString();

		return getSetting(section, parameter);
	}

	/**
	 * Returns the value of the parameter in the given section.
	 * 
	 * @param section	Section of the settings to look up.
	 * @param parameter Parameter to look up.
	 * @return The value of the setting at the given section and parameter, or
	 * 		   <code>null</code> if no such parameter exists.
	 */
	private static String getSetting(String section, String parameter) {

		String result = null;

		// Look first to the override settings
		if (newsettings.containsKey(section)) {
			HashMap<String, String> sectiontest = newsettings.get(section);

			if (sectiontest.containsKey(parameter)) {
				result = sectiontest.get(parameter);
			}
		}
		// If nothing, look to the file
		if (result == null) {
			result = settings.getValue(section, parameter);
		}

		return result;
	}

	/**
	 * Loads and applies the settings of the settings file, RedHorizon.ini.
	 */
	public static void init() {

		settings = new IniFile("RedHorizon.ini");
		((IniFile)settings).readFromFile();
	}

	/**
	 * Saves the values of all settings, old or updated, to the RedHorizon.ini
	 * file.
	 */
	private static void saveSettings() {

		// Only perform a save if there have been changes
		if (newsettings.size() == 0) {
			return;
		}

		// Otherwise, replace any old values with new ones
		String allsettings = settings.toString();
		String[] lines = allsettings.split("\n");
		StringBuilder result  = new StringBuilder(allsettings);
		String section = null;
		String parameter = null;

		for (String line: lines) {
			line = line.trim();

			// Skip empty and comment lines
			if (line.length() == 0 || line.charAt(0) == ';') {
				continue;
			}

			// Update to current section
			if (line.charAt(0) == '[') {
				section = line;
				continue;
			}
			// Update to current parameter
			if (section != null) {
				parameter = line.substring(0, line.indexOf("=")).trim();
			}

			// Override changed values
			String original = settings.getValue(section, parameter);
			String override = getSetting(section, parameter);
			if (!override.equals(original)) {
				int start = result.indexOf(line);
				int end   = line.contains(";") ? result.indexOf("\t", start) : result.indexOf("\r\n", start);
				result.replace(start, end, parameter + " = " + override);
			}
		}

		// Write the updated settings
		FileUtility.writeTextFile("RedHorizon.ini", CharBuffer.wrap(result));
	}

	/**
	 * Alters the value of a value in the settings file.  This doesn't actually
	 * alter the value in the file or the file representation itself, but
	 * instead saves the new setting to another collection.  This collection
	 * will be checked first whenever a setting lookup or the file save is
	 * performed.
	 * 
	 * @param settingkey The parameter to alter.
	 * @param newvalue	The new value to assign the parameter.
	 */
	public static void setSetting(Enum<?> settingkey, String newvalue) {

		String section   = settingkey.getClass().getSimpleName();
		section = "[" + section.substring(0, section.indexOf("Settings")) + "]";
		String parameter = settingkey.toString();

		HashMap<String, String> newsection;
		if (newsettings.containsKey(section)) {
			newsection = newsettings.get(section);
		}
		else {
			newsection = new HashMap<String, String>();
		}
		newsection.put(parameter, newvalue);
		newsettings.put(section, newsection);
	}
}
