
// ================================================
// Scanner's Java - Locale-specific program strings
// ================================================

package redhorizon.strings;

import redhorizon.settings.CommonSettings;
import redhorizon.settings.Settings;

import java.text.NumberFormat;
import java.util.HashMap;
import java.util.Locale;
import java.util.ResourceBundle;

/**
 * Class to access all locale-specific program strings (such as text for the
 * interface, text for in-game, error messages, etc).  The different text for
 * each localization can be found under the localizations source directory.
 * 
 * @author Emanuel Rabina
 */
public class ProgramStrings {

	private static Locale locale;
	private static HashMap<String,ResourceBundle> messagebundles = new HashMap<String,ResourceBundle>();
	private static NumberFormat percentformatter;

	/**
	 * Hidden default constructor as this class is only ever meant to be used
	 * statically.
	 */
	private ProgramStrings() {
	}

	/**
	 * Adds a <code>ResourceBundle</code> containing (locale-specific) message
	 * text.  The text within the bundles can then be retrieved using the
	 * {@link #getText(Enum, String...)} method.  The name of the bundle that
	 * will be loaded must have a filename similar to classname.
	 *
	 * @param keytype	 The <code>Enum</code> that will be used to access the
	 * 					 messages within the bundle.
	 */
	public static void addMessageBundle(Class<?> keytype) {

		String bundlename = keytype.getSimpleName();
		messagebundles.put(bundlename, ResourceBundle.getBundle(bundlename, locale));
	}

	/**
	 * Converts the floating-point value into a percentage string.
	 *
	 * @param floatval The floating-point number.
	 * @return The value as a percentage, in the user's selected locale.
	 */
	public static String floatToPercent(float floatval) {

		return percentformatter.format(floatval);
	}

	/**
	 * Returns localized text from the <code>ResourceBundle</code> which was
	 * added with the given enumerated type.
	 *
	 * @param textkey Key to the text in the resource bundle.
	 * @param strings List of replacement strings for the {n} values in the
	 * 				  message text.
	 * @return Localized text for the given key.
	 */
	public static String getText(Enum<?> textkey, String... strings) {

		return getText(textkey.getClass().getSimpleName(), textkey.name(), strings);
	}

	/**
	 * Returns the piece of localized text associated with the enum type key.
	 *
	 * @param keytype The <code>ResourceBundle</code> which has this text.
	 * @param textkey Enum type which matches the text to be displayed.
	 * @return Localized text for the given key.
	 */
	private static String getText(String keytype, String textkey) {

		return messagebundles.get(keytype).getString(textkey);
	}

	/**
	 * Returns the piece of localized text associated with the enum type key,
	 * using the replacement text provided in the <code>String</code> array.
	 * Use this method for messages in the properties file with {x} values to be
	 * replaced by their counterparts in the array.
	 *
	 * @param keytype The enumerated type used to store the
	 * 				  <code>ResourceBundle</code>.
	 * @param textkey Enum type which matches the text to be displayed.
	 * @param strings Array of replacement strings for the {n} values in the
	 * 				   properties file.
	 * @return Localized text for the given key.
	 */
	private static String getText(String keytype, String textkey, String... strings) {

		// NOTE: Standard property substitution in Java doesn't work the way I
		//		 want.  The below fixes that.
		StringBuilder text = new StringBuilder(getText(keytype, textkey));
		for (int i = 0; i < strings.length; i++) {
			while (true) {
				int nextparam = text.indexOf("{" + i + "}");
				if (nextparam == -1) {
					break;
				}
				text.replace(nextparam, nextparam + 3, strings[i]);
			}
		}
		return text.toString();
	}

	/**
	 * Sets the locale to use for the resource bundles which will be loaded.
	 */
	public static void init() {

		String localestring = Settings.getSetting(CommonSettings.LOCALE);
		if (localestring.contains("_")) {
			String language = localestring.substring(0, localestring.indexOf('_'));
			String country  = localestring.substring(localestring.indexOf('_') + 1);
			locale = new Locale(language, country);
		}
		else {
			locale = new Locale(localestring);
		}

		// Create appropriate number formatters
		percentformatter = NumberFormat.getPercentInstance(locale);

		// Load all redhorizon-common bundles
		addMessageBundle(Common_Filetypes.class);
	}

	/**
	 * Converts the percentage-string into a floating-point value.
	 *
	 * @param percentval Locale-specific percentage value.
	 * @return The percentage as a floating-point value.
	 */
	public static float percentToFloat(String percentval) {

		return percentformatter.parse(percentval).floatValue();
	}
}
