
// ========================================
// Scanner's Java - Buffer creation utility
// ========================================

package redhorizon.utilities;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;

/**
 * Buffer utility class with the only purpose of creating new buffers with
 * native byte ordering.
 * 
 * @author Emanuel Rabina.
 */
public class BufferUtility {

	/**
	 * Hidden default constructor, as this class is only ever meant to be used
	 * statically.
	 */
	private BufferUtility() {
	}

	/**
	 * Creates a new non-direct <code>ByteBuffer</code> with the specified
	 * capacity.
	 * 
	 * @param capacity Number of bytes the buffer should be capable of holding.
	 * @return New <code>ByteBuffer</code>.
	 */
	public static ByteBuffer newByteBuffer(int capacity) {

		return newByteBuffer(capacity, false);
	}

	/**
	 * Creates a new <code>ByteBuffer</code> with the specified capacity.  Can
	 * either be direct or non-direct.
	 * 
	 * @param capacity Number of bytes the buffer should be capable of holding.
	 * @param isdirect <code>true</code> if a direct <code>ByteBuffer</code> is
	 * 				   requested.
	 * @return New <code>ByteBuffer</code>.
	 */
	public static ByteBuffer newByteBuffer(int capacity, boolean isdirect) {

		ByteBuffer bytebuffer;
		if (isdirect) {
			bytebuffer = ByteBuffer.allocateDirect(capacity).order(ByteOrder.nativeOrder());
		}
		else {
			bytebuffer = ByteBuffer.allocate(capacity).order(ByteOrder.nativeOrder());
		}
		return bytebuffer;
	}

	/**
	 * Creates a new <code>ByteBuffer</code> by wrapping an existing
	 * <code>byte</code>[].
	 * 
	 * @param bytes Array of bytes.
	 * @return New <code>ByteBuffer</code> backed by the given byte array.
	 */
	public static ByteBuffer wrapByteBuffer(byte[] bytes) {

		return ByteBuffer.wrap(bytes).order(ByteOrder.nativeOrder());
	}
}
