
// ======================================
// Scanner's Java - File format utilities
// ======================================

package redhorizon.utilities;

import java.nio.ByteBuffer;

/**
 * Utility class with methods to deal with the several file formats used by both
 * Red Alert and Tiberium Dawn.  The methods here are mainly used to decode,
 * decrypt, or decompress the files used, as well as conversion to turn the new
 * raw data into forms more easily used by OpenGL, and OpenAL.<br>
 * <br>
 * Much of the C&C-specific methods (Format20, Format40, Format80,
 * 16bitIMAADPCM) have been adapted from code written by Vladan Bato.  Credit
 * also goes to Asatur V. Nazarian for the 8bitWSADPCM code.
 * 
 * @author Emanuel Rabina
 */
public class CNCCodec {

	// Base-64 table as used by the *pack sections of RA's maps
	private static final String base64 =
		"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

	// WS-ADPCM 2-bit adjustment table
	private static final int[] wstable2bit = {
		-2, -1, 0, 1
	};

	// WS-ADPCM 4-bit adjustment table
	private static final int[] wstable4bit = {
		-9, -8, -6, -5, -4, -3, -2, -1, 0,  1,  2,  3,  4,  5,  6,  8
	};

	// IMA-ADPCM adjustment table
	private static final int[] imaadjusttable = {
		-1, -1, -1, -1, 2, 4, 6, 8
	};

	// IMA-ADPCM step table
	private static final int[] imasteptable = {
		    7,     8,     9,    10,    11,    12,     13,    14,    16,
		   17,    19,    21,    23,    25,    28,     31,    34,    37,
		   41,    45,    50,    55,    60,    66,     73,    80,    88,
		   97,   107,   118,   130,   143,   157,    173,   190,   209,
		  230,   253,   279,   307,   337,   371,    408,   449,   494,
		  544,   598,   658,   724,   796,   876,    963,  1060,  1166,
		 1282,  1411,  1552,  1707,  1878,  2066,   2272,  2499,  2749,
		 3024,  3327,  3660,  4026,  4428,  4871,   5358,  5894,  6484,
		 7132,  7845,  8630,  9493, 10442, 11487,  12635, 13899, 15289,
		16818, 18500, 20350, 22385, 24623, 27086,  29794, 32767
	};

	/**
	 * Hidden default constructor, as this class is only ever meant to be used
	 * statically.
	 */
	private CNCCodec() {
	}

	/**
	 * Decodes base64 data into standard binary.
	 * 
	 * @param source Original encoded data.
	 * @param dest	 Buffer to store the decoded data.
	 */
	public static void decodeBase64(ByteBuffer source, ByteBuffer dest) {

		// Find out how many bits to cut from the end
		int cullbits = 0;
		for (int i = source.capacity() - 1; source.get(i) == (byte)'='; i--) {
			cullbits += 2;
			source.limit(source.limit() - 1);
		}

		// Construct the binary values
		for (int i = 0; source.limit() - i >= 4; i += 4) {
			byte a = decodeBase64CharVal(source.get());
			byte b = decodeBase64CharVal(source.get());
			byte c = decodeBase64CharVal(source.get());
			byte d = decodeBase64CharVal(source.get());

			dest.put((byte)(a << 2 | b >>> 4));
			dest.put((byte)(b << 4 | c >>> 2));
			dest.put((byte)(c << 6 | d));
		}

		// Pick-up any leftovers
		int diff = source.limit() - source.position();
		if (cullbits == 4 && diff == 2) {
			byte a = decodeBase64CharVal(source.get());
			byte b = decodeBase64CharVal(source.get());

			dest.put((byte)(a << 2 | b >>> 4));
		}
		else if (cullbits == 2 && diff == 3) {
			byte a = decodeBase64CharVal(source.get());
			byte b = decodeBase64CharVal(source.get());
			byte c = decodeBase64CharVal(source.get());

			dest.put((byte)(a << 2 | b >>> 4));
			dest.put((byte)(b << 4 | c >>> 2));
		}
		dest.limit(dest.position()).rewind();
	}

	/**
	 * Gets the Base-64 value assigned to the characters used in the table.
	 * 
	 * @param charval The character whose value is to be found.
	 * @return The value assigned to that character.
	 */
	private static byte decodeBase64CharVal(byte charval) {

		byte value = (byte)0x80;
		for (int i = 0; i < base64.length(); i++) {
			if (charval == (byte)base64.charAt(i)) {
				value = (byte)i;
				break;
			}
		}
		return value;
	}

	/**
	 * Decodes a Format2 SHP image.<br>
	 * <br>
	 * A Format2 image is used in the Dune 2 SHP files and is a simple
	 * compression technique which can be decompressed as follows:<br>
	 * 1) 0 c = Fill the next c bytes with 0.<br>
	 * 2) v = Write v.<br>
	 * 
	 * @param source Original encoded image bytes.
	 * @param dest	 Buffer to store the decoded image bytes.
	 */
	public static void decodeFormat2(ByteBuffer source, ByteBuffer dest) {

		while (source.hasRemaining()) {
			byte command = source.get();

			// Fill 0s
			if (command == 0) {
				int count = source.get() & 0xff;
				while (count-- > 0) {
					dest.put((byte)0);
				}
			}
			// Write direct value
			else {
				dest.put(command);
			}
		}
		dest.limit(dest.position()).rewind();
	}

	/**
	 * Decodes a Format20 SHP image.<br>
	 * <br>
	 * A Format20 image is just a Format40 image which is based on the image
	 * data that came just before the current one.  This method is merely a call
	 * to the Format40 decoding method, but kept so that the differences
	 * between format types can be upheld.
	 * 
	 * @param source Original encoded image bytes.
	 * @param dest	 Buffer to store the decoded image bytes.
	 */
	public static void decodeFormat20(ByteBuffer source, ByteBuffer dest) {

		decodeFormat40(source, dest);
	}

	/**
	 * Decodes a Format40 SHP image.<br>
	 * <br>
	 * The documentation of the Format40 SHP image defines 6 special commands
	 * used depending upon the head bits of the byte read (much like machine
	 * instructions).  Using a notation found in XCCU, they are as follows:<br>
	 * 1) 00000000 c v = Xor next c bytes with v.<br>
	 * 2) 0ccccccc = Xor the next c bytes from source with those in dest.<br>
	 * 3) 10000000 0c c = Skip the next c bytes.<br>
	 * 4) 10000000 10c c = Xor the next c bytes from source with those in dest.<br>
	 * 5) 10000000 11c c v = Xor the next c bytes with v.<br>
	 * 6) 1ccccccc = Skip the next c bytes.
	 * 
	 * @param source Original encoded image bytes.
	 * @param dest	 Buffer to store the decoded image bytes.
	 */
	public static void decodeFormat40(ByteBuffer source, ByteBuffer dest) {

		int count;

		while (true) {
			byte command = source.get();

			// b7 = 0
			if ((command & 0x80) == 0) {

				// Command #1
				if (command == 0) {
					count = source.get() & 0xff;
					byte fill = source.get();
					while (count-- > 0) {
						dest.put((byte)(dest.get(dest.position()) ^ fill));
					}
				}
				// Command #2
				else {
					count = command;
					while (count-- > 0) {
						dest.put((byte)(dest.get(dest.position()) ^ source.get()));
					}
				}
			}
			// b7 = 1
			else {
				count = command & 0x7f;

				// b6-0 = 0
				if (count == 0) {
					count = source.getShort() & 0xffff;
					command = (byte)(count >>> 8);

					// b7 of next byte = 0
					if ((command & 0x80) == 0) {

						// Finished decoding
						if (count == 0) {
							break;
						}

						// Command #3
						dest.position(dest.position() + count);
					}
					// b7 of next byte = 1
					else {
						count &= 0x3fff;

						// Command #4
						if ((command & 0x40) == 0) {
							while (count-- > 0) {
								dest.put((byte)(dest.get(dest.position()) ^ source.get()));
							}
						}
						// Command #5
						else {
							byte fill = source.get();
							while (count-- > 0) {
								dest.put((byte)(dest.get(dest.position()) ^ fill));
							}
						}
					}
				}
				// b6-0 != 0
				else {

					// Command #6
					dest.position(dest.position() + count);
				}
			}
		}
		dest.limit(dest.position()).rewind();
	}

	/**
	 * Decodes a Format80 SHP image.<br>
	 * <br>
	 * The documentation of the Format80 SHP image defines 5 special commands
	 * used depending upon the head bits of the byte read (much like machine
	 * instructions).  Using a notation found in XCCU, they are as follows:<br>
	 * 1) 0cccpppp p = Copy c + 3 bytes from dest.pos - p to dest.pos.<br>
	 * 2) 10cccccc = Copy next c bytes from source to dest.<br>
	 * 3) 11cccccc p p = Copy c bytes from p.<br>
	 * 4) 11111110 c c v = Write c bytes with v.<br>
	 * 5) 11111111 c c p p = Copy c bytes from p.
	 * 
	 * @param source Original compressed image bytes.
	 * @param dest	 Buffer to store the uncompressed image bytes.
	 */
	public static void decodeFormat80(ByteBuffer source, ByteBuffer dest) {

		int count;
		int copypos;

		while (true) {
			byte command = source.get();

			// b7 = 0
			if ((command & 0x80) == 0) {

				// Command #1
				count = (command >>> 4) + 3;
				copypos = ((command & 0x0f) << 8) | (source.get() & 0xff);
				copypos = dest.position() - copypos;
				while (count-- > 0) {
					dest.put(dest.get(copypos++));
				}
			}
			// b7 = 1
			else {
				count = command & 0x3f;

				// b6 = 0
				if ((command & 0x40) == 0) {

					// Finished decoding
					if (count == 0) {
						break;
					}

					// Command #2
					while (count-- > 0) {
						dest.put(source.get());
					}
				}
				// b6 = 1
				else {

					// Command #3
					if (count < 0x3e) {
						count += 3;
						copypos = source.getShort() & 0xffff;
						while (count-- > 0) {
							dest.put(dest.get(copypos++));
						}
					}
					// Command #4
					else if (count == 0x3e) {
						count = source.getShort() & 0xffff;
						byte fill = source.get();
						while (count-- > 0) {
							dest.put(fill);
						}
					}
					// Command #5
					else {
						count = source.getShort() & 0xffff;
						copypos = source.getShort() & 0xffff;
						while (count-- > 0) {
							dest.put(dest.get(copypos++));
						}
					}
				}
			}
		}
		dest.limit(dest.position()).rewind();
	}

	/**
	 * Decodes the [MapPack] section of a scenario file, converting the base-64
	 * data to binary, and then decompressing it to the full map size (48k).
	 * 
	 * @param source Original encoded data as per the characters in the map
	 * 				 file.  These chars are intended to be 8-bit, instead of
	 * 				 Java's 16-bit chars, so conversion will need to be done
	 * 				 before-hand.
	 * @param dest	 48k destination buffer containing full uncompressed binary
	 * 				 data on the [MapPack] section.
	 */
	public static void decodeMapPack(ByteBuffer source, ByteBuffer dest) {

		// Decode standard pack data with size of 6 chunks
		decodePackData(source, dest, 6);
	}

	/**
	 * Decodes the [OverlayPack] section of a scenario file, converting the
	 * base-64 data to binary, and then decompressing it to the full overlay
	 * size (16k).
	 * 
	 * @param source Original encoded data as per the characters in the map
	 * 				 file.  These chars are intended to be 8-bit, instead of
	 * 				 Java's 16-bit chars, so conversion will need to be done
	 * 				 before-hand.
	 * @param dest	 16k destination buffer containing full uncompressed binary
	 * 				 data on the [OverlayPack] section.
	 */
	public static void decodeOverlayPack(ByteBuffer source, ByteBuffer dest) {

		// Decode standard pack data with size of 2 chunks
		decodePackData(source, dest, 2);
	}

	/**
	 * Generic [*Pack] decoding routines, used for [MapPack] and [OverlayPack]
	 * sections.
	 * 
	 * @param source Original encoded data.
	 * @param dest	 Destination buffer for decoded data.
	 * @param chunks The number of chunks inside the [*Pack] section.
	 */
	private static void decodePackData(ByteBuffer source, ByteBuffer dest, int chunks) {

		// Decode base64 data into standard binary
		ByteBuffer mapbytes2 = BufferUtility.newByteBuffer(8192 * chunks);
		CNCCodec.decodeBase64(source, mapbytes2);

		// Decode pack data, 'chunks' number of chunks
		ByteBuffer[] mapchunks = new ByteBuffer[chunks];
		for (int i = 0; i < chunks; i++) {
			mapchunks[i] = BufferUtility.newByteBuffer(8192);
		}

		for (int i = 0, pos = 0; i < chunks; i++) {

			// Get following chunk size, skip 0x20 (unknown) byte
			int a = mapbytes2.get() & 0xff;
			int b = mapbytes2.get() & 0xff;
			int c = mapbytes2.get() & 0xff;
					mapbytes2.position(mapbytes2.position() + 1);
			int chunksize = (c << 16) | (b << 8) | a;

			// Decode that chunk, put it into one of the buffers in the array
			CNCCodec.decodeFormat80(mapbytes2, mapchunks[i]);
			pos += 4 + chunksize;
			mapbytes2.position(pos);
		}

		// Collate chunks into total map data
		for (ByteBuffer mapchunk: mapchunks) {
			dest.put(mapchunk);
		}
		dest.rewind();
	}

	/**
	 * Decodes a section of data which has been encoded using the run-length
	 * encoding scheme.  The 67 variant of RLE is one where the count byte is
	 * denoted by bits 6 & 7 set to 1.
	 * 
	 * @param source Original encoded data.
	 * @param dest	 Buffer to store the uncompressed data.
	 */
	public static void decodeRLE67(ByteBuffer source, ByteBuffer dest) {

		// Until the source is exhausted
		while (source.hasRemaining()) {
			byte nextbyte = source.get();

			// Count byte & copy byte run
			if ((nextbyte & 0xc0) == 0xc0) {
				int count = nextbyte & 0x3f;
				byte copy = source.get();

				while (count-- > 0) {
					dest.put(copy);
				}
			}
			// Non-count byte
			else {
				dest.put(nextbyte);
			}
		}
		dest.rewind();
	}

	/**
	 * Decompresses AUD file data using Westwood's proprietary 8-bit ADPCM
	 * decompression.  This is used for very few sound samples in both Red Alert
	 * and Tiberium Dawn, most notably the infantry death sounds.<br>
	 * <br>
	 * This decompression technique is only for mono sound data.<br>
	 * <br>
	 * Note: does this method even get used?  I haven't found a single instance
	 *       of an 8-bit file.
	 * 
	 * @param source Original compressed audio data.
	 * @param dest	 Buffer to store the uncompressed data.
	 */
	public static void decompress8bitWSADPCM(ByteBuffer source, ByteBuffer dest) {

		// Mere copy if chunk is uncompressed
		if (source.capacity() == dest.capacity()) {
			dest.put(source).rewind();
			return;
		}

		// Decompression
		short sample = 0x80;
		while (dest.hasRemaining()) {

			short input = source.getShort();
			input <<= 2;
			byte command = (byte)(input >>> 8);
			byte count   = (byte)((input & 0xff) >> 2);

			// No compression
			if (command == 2) {

				if ((count & 0x20) != 0) {
					count <<= 3;
					sample += count >> 3;
					dest.put((byte)sample);
				}
				else {
					for ( ; count >= 0; count--) {
						dest.put(source.get());
					}
					sample = dest.get(dest.position() - 1);
					sample &= 0xffff;
				}
			}

			// 2x compression (4-bit -> 8-bit)
			else if (command == 1) {

				for ( ; count >= 0; count--) {
					command = source.get();

					sample += wstable4bit[command & 0x0f];
					sample = (short)Math.min((short)Math.max(sample, 0), 255);
					dest.put((byte)sample);

					sample += wstable4bit[command >>> 4];
					sample = (short)Math.min((short)Math.max(sample, 0), 255);
					dest.put((byte)sample);
				}
			}

			// 4x compression (2-bit -> 8-bit)
			else if (command == 0) {

				for ( ; count >= 0; count--) {
					command = source.get();

					sample += wstable2bit[command & 0x03];
					sample = (short)Math.min((short)Math.max(sample, 0), 255);
					dest.put((byte)sample);

					sample += wstable2bit[(command >>> 2) & 0x03];
					sample = (short)Math.min((short)Math.max(sample, 0), 255);
					dest.put((byte)sample);

					sample += wstable2bit[(command >>> 4) & 0x03];
					sample = (short)Math.min((short)Math.max(sample, 0), 255);
					dest.put((byte)sample);

					sample += wstable2bit[(command >>> 6) & 0x03];
					sample = (short)Math.min((short)Math.max(sample, 0), 255);
					dest.put((byte)sample);
				}
			}

			// Straight copy
			else {
				while (count-- >= 0) {
					dest.put((byte)sample);
				}
			}
		}
		dest.limit(dest.position()).rewind();
	}

	/**
	 * Decompresses AUD file data using IMA-ADPCM decompression.  This
	 * decompression technique is used with Red Alert's and Tiberium Dawn's
	 * 16-bit audio files.  For the 8-bit, a proprietary format is used and can
	 * be decoded/decompressed using
	 * {@link #decompress8bitWSADPCM(ByteBuffer,ByteBuffer)}.
	 * 
	 * @param source Original compressed audio data.
	 * @param dest	 Buffer to store the uncompressed data.
	 * @param update 2-<code>int</code> array, containing the latest index and
	 * 				 sample values respectively.
	 */
	public static void decompress16bitIMAADPCM(ByteBuffer source, ByteBuffer dest, int[] update) {

		int index  = update[0];
		int sample = update[1];

		// Until all the compressed data has been decompressed
		for (int sampleindex = 0; sampleindex < source.limit() << 1; sampleindex++) {

			// The 4-bit command
			byte code = source.get(sampleindex >> 1);
			code = (sampleindex % 2 == 1) ? (byte)(code >>> 4) : (byte)(code & 0x0f);

			int step = imasteptable[index];
			int delta = step >>> 3;

			// Expansion of the multiplication in the original pseudo code
			if ((code & 0x01) != 0) {
				delta += step >>> 2;
			}
			if ((code & 0x02) != 0) {
				delta += step >>> 1;
			}
			if ((code & 0x04) != 0) {
				delta += step;
			}

			// Sign bit = 1
			if ((code & 0x08) != 0) {
				sample -= delta;
				sample = Math.max(sample, -32768);
			}
			// Sign bit = 0
			else {
				sample += delta;
				sample = Math.min(sample, 32767);
			}

			// Save result to destination buffer
			dest.putShort((short)sample);

			// Index/Step adjustments
			index += imaadjusttable[code & 0x07];
			index = Math.min(Math.max(index, 0), 88);
		}
		dest.limit(dest.position()).rewind();
		update[0] = index;
		update[1] = sample;
	}

	/**
	 * Encodes a series of bytes to the using the Format2 algorithm.<br>
	 * <br>
	 * A Format2 image is used in the Dune 2 SHP files and is a simple
	 * compression technique which can be compressed as follows:<br>
	 * 1) Series of 0s = write a 0 followed by how many bytes of 0 there were. A
	 * 					 maximum of <code>slices</code> bytes can be described
	 * 					 by this method.<br>
	 * 2) Any non-0 byte = write that non-0 byte.
	 *
	 * @param source Original raw image bytes.
	 * @param slices Number of rows to run the encoding on.  This is usually the
	 * 				 height of the original image.
	 * @param dest   Buffer to store the encoded image bytes.
	 */
	public static void encodeFormat2(ByteBuffer source, int slices, ByteBuffer dest) {

		int count;
		int limit = source.limit() / slices;

		while (source.hasRemaining()) {
			count = 0;

			for (int pixel = 0; pixel < limit; pixel++) {
				byte value = source.get();

				// Read 0
				if (value == 0) {
					count++;
				}
				// Read non-0
				else {

					// End of series of 0s, describe series
					if (count > 0) {
						dest.put(new byte[]{ 0, (byte)count });
						count = 0;
					}
					// Write non-0 value
					dest.put(value);
				}
			}
			// Finish remaining series of 0s (if any)
			if (count > 0) {
				dest.put(new byte[]{ 0, (byte)count });
			}
		}
		source.rewind();
		dest.limit(dest.position()).rewind();
	}
}
