
// ==============================================
// Scanner's Java - Input end of a streaming pipe
// ==============================================

package redhorizon.utilities;


import java.nio.ByteBuffer;
import java.util.concurrent.BlockingQueue;

/**
 * The input end of a data stream, from which data can be pushed into, to be
 * read by the source end at a later time.
 * 
 * @author Emanuel Rabina
 */
public class DataSink {

	private final DataStream pipe;
	private final BlockingQueue<ByteBuffer[]> sink;

	/**
	 * Package-only constructor, creates a new input stream using the given
	 * <code>BlockingQueue</code> implementation, using it's 'add' methods for
	 * writing.
	 * 
	 * @param pipe The master pipe connection.
	 * @param sink The input end of a pipe.
	 */
	DataSink(DataStream pipe, BlockingQueue<ByteBuffer[]> sink) {

		this.pipe = pipe;
		this.sink = sink;
	}

	/**
	 * Indicates that there is no more data to write.  Headers indicating that
	 * the following (non-existant) data chunk is -1 bytes, are written.  This
	 * is picked-up by the reader end as a notification that there is no more
	 * data to read.
	 */
	public void finish() {

		// Generate negative header
		ByteBuffer header = BufferUtility.newByteBuffer(pipe.chunksize << 2);
		for (int i = 0; i < pipe.chunksize; i++) {
			header.putInt(-1);
		}
		header.rewind();

		ByteBuffer[] packet = { header };

		// Write negative header
		sink.put(packet);
	}

	/**
	 * Writes an array of <code>ByteBuffer</code>s to this end of the stream.
	 * This method will block if the maximum number of chunks currently in this
	 * pipe has been reached, and will not finish until there is some room to
	 * write the data to, or the pipe has been closed.<br>
	 * <br>
	 * Note that what constitues as 1 chunk is dependant of the use of this
	 * pipe.  If it's for 1 or either audio/video data, then 1 chunk =
	 * <code>int</code> of the size of the following <code>ByteBuffer</code> +
	 * the <code>ByteBuffer</code> itself.  If it's both audio & video data,
	 * then 1 chunk = <code>int</code> x 2 + <code>ByteBuffer</code> +
	 * <code>ByteBuffer</code>.
	 * 
	 * @param data <code>ByteBuffer</code>[] of data to write to the stream.
	 * 			   The number of elements in this array must equal the number of
	 * 			   <code>ByteBuffer</code>s that constitute a chunk.
	 */
	public void write(ByteBuffer[] data) {

		// Generate header information
		ByteBuffer[] packet = new ByteBuffer[pipe.chunksize + 1];
		ByteBuffer header = BufferUtility.newByteBuffer(pipe.chunksize << 2);
		for (int i = 0; i < pipe.chunksize; i++) {
			header.putInt(data[i] != null ? data[i].limit() : 0);
		}
		header.rewind();

		// Combine data, write packet
		packet[0] = header;
		for (int i = 0; i < pipe.chunksize; i++) {
			packet[i + 1] = data[i];
		}
		sink.put(packet);
	}
}
