
// ===============================================
// Scanner's Java - Output end of a streaming pipe
// ===============================================

package redhorizon.utilities;

import java.nio.ByteBuffer;
import java.util.concurrent.BlockingQueue;

/**
 * The output end of a data stream, from which data can be read out of, once
 * something is available to be read.
 * 
 * @author Emanuel Rabina
 */
public class DataSource {

	private final DataStream pipe;
	private final BlockingQueue<ByteBuffer[]> source;

	/**
	 * Package-only constructor, creates a new output stream using the given
	 * <code>BlockingQueue</code> implementation.
	 * 
	 * @param pipe	   The master pipe connection.
	 * @param source   The output end of a pipe.
	 */
	DataSource(DataStream pipe, BlockingQueue<ByteBuffer[]> source) {

		this.pipe   = pipe;
		this.source = source;
	}

	/**
	 * Reads from the stream, the next data chunk, returning it into the given
	 * array of <code>ByteBuffer</code> objects.  The number of elements in the
	 * returned array will match the number of <code>ByteBuffer</code> objects
	 * that constitutes a data chunk, defined at the constructor of the master
	 * {@link DataStream}.
	 * 
	 * @param data Empty <code>ByteBuffer</code>[] to hold the incoming data.
	 * @return <code>true</code> if all of the buffers in the array were
	 * 		   allocated, <code>false</code> otherwise.
	 */
	public boolean read(ByteBuffer[] data) {

		// Read header information
		ByteBuffer[] packet = source.take();
		ByteBuffer header = packet[0];

		// Quit if it's all less-than 0 (meaning streaming operation complete)
		int total = 0;
		for (int i = 0; i < pipe.chunksize; i++) {
			total += header.getInt(i);
		}
		if (total < 0) {
			return false;
		}

		// Read data from queue
		for (int i = 0; i < pipe.chunksize; i++) {
			data[i] = packet[i + 1];
		}
		return true;
	}
}
