
// ===========================================
// Scanner's Java - Pipe/Stream implementation
// ===========================================

package redhorizon.utilities;

import java.nio.ByteBuffer;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

/**
 * Wrapper of a <code>BlockingQueue</code> type of Java's, this class provides
 * several methods so that the underlying queue can be used as the
 * inputs/outputs of streaming file data.<br>
 * <br>
 * The format of the data on the queue is as follows:<br>
 * <li>int - size of the corresponding data chunk</li>
 * <li>int <code>1..n</code></li>
 * <li>ByteBuffer - the corresponding data chunk</li>
 * <li>ByteBuffer <code>1..n</code></li>
 * <br>
 * The value of <code>n</code> is determined at the constructor of this stream.
 * If this stream contains only 1 type of data (eg: pictures only, or sound
 * only), then n = 1.  If there are 2 types of data (eg: both pictures and
 * sound), then n = 2.  As for the order of the data, that is up to the
 * implementations to decide.<br>
 * <br>
 * Once the end of the stream writer has been reached (when both audio and video
 * chunk sizes are reported as negative, then the end of the stream has been
 * reached.
 * 
 * @author Emanuel Rabina
 */
public class DataStream {

	// Number of chunks to hold in the stream before blocking
	private static final int STREAM_BUFFER_SIZE = 10;

	private final BlockingQueue<ByteBuffer[]> streambuffer;
	private final DataSink sink;
	private final DataSource source;
	final int chunksize;

	private Thread decoder;

	/**
	 * Constructor, opens a new blocking data stream configured with the given
	 * parameters.
	 * 
	 * @param chunksize The number of <code>ByteBuffer</code>s that constitute a
	 * 					'chunk' of data. (see the class description for a
	 * 					definition of a chunk).
	 */
	public DataStream(int chunksize) {

		streambuffer = new LinkedBlockingQueue<ByteBuffer[]>(STREAM_BUFFER_SIZE);
		sink    = new DataSink(this, streambuffer);
		source  = new DataSource(this, streambuffer);

		this.chunksize = chunksize;
	}

	/**
	 * Returns the sink/input end of this stream.
	 * 
	 * @return The <code>DataStreamSink</code> for this stream.
	 */
	public DataSink getSink() {

		return sink;
	}

	/**
	 * Returns the source/output end of this stream.
	 * 
	 * @return The <code>DataStreamSource</code> for this stream.
	 */
	public DataSource getSource() {

		return source;
	}

	/**
	 * Returns whether or not the stream has been started.
	 * 
	 * @return <code>true</code> if the stream has been started,
	 * 		   <code>false</code> if it hasn't.
	 */
	public boolean isStreaming() {

		return decoder.isAlive();
	}

	/**
	 * Sets the streaming thread to use the given <code>Runnable</code>
	 * implementation.
	 * 
	 * @param algorithm <code>ControlledThread</code> whose <code>run()</code>
	 * 					method incorporates the necessary methods to decode data
	 * 					from a file down the stream.
	 */
	public void setAlgorithm(Runnable algorithm) {

		decoder = new Thread(algorithm);
	}

	/**
	 * Called externally and indicates to the stream to start the
	 * <code>Thread</code> of the streaming algorithm registered in this
	 * class.<br>
	 * <br>
	 * The algorithm is defined in the argument passed to the
	 * {@link #setAlgorithm(Runnable)} method.
	 */
	public void startStream() {

		decoder.start();
	}
}
