
// ===========================================
// Scanner's Java - File channel utility class
// ===========================================

package redhorizon.utilities;

import java.io.FileDescriptor;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;

/**
 * Essentially a wrapper of a <code>RandomAccessFile</code>, this class provides
 * several utility methods for actions involving something along the lines of a
 * pointer to a location in a file.<br>
 * <br>
 * It's main use would be for being able to move around either an entire file or
 * just a section of a file, much like a <code>RandomAccessFile</code>, except
 * that this implementation contains lower and upper limits to keep it localized
 * to a smaller area (such as a file within a file).  From the point-of-view of
 * any classes that use the <code>FilePointer</code>, these bounds maintain the
 * illusion that it is a whole file.<br>
 * <br>
 * Note that this implementation uses <code>int</code>-sized values for things
 * such as file size and file position.  This restricts it to file size of about
 * 4GBs.  If more is needed, expand the <code>int</code>s to <code>long</code>s.
 * 
 * @author Emanuel Rabina
 */
public class FilePointer {

	private RandomAccessFile file;
	private int lowerlimit;
	private int upperlimit;

	/**
	 * Constructor, uses the given name to generate a read-only
	 * <code>FilePointer</code> from.  Lower and upper bounds of the range of
	 * the pointer are 0 and the size of the file respectively.
	 * 
	 * @param filename Name of the file on the file system.
	 */
	public FilePointer(String filename) {

		this(filename, true);
	}

	/**
	 * Constructor, uses the given name to generate a <code>FilePointer</code>
	 * in the specified read/write mode.  Lower and upper bounds of the range of
	 * the pointer are 0 and the size of the file respectively.
	 * 
	 * @param filename Name of the file on the file system.
	 * @param readonly Whether to allow writes to this file.  Specify
	 * 				   <code>false</code> for file writing.
	 */
	public FilePointer(String filename, boolean readonly) {

		// Use existing file or create new blank file
		if (readonly) {
			file = new RandomAccessFile(filename, "r");
		}
		else {
			file = new RandomAccessFile(filename, "rw");
			file.setLength(0);
		}
		lowerlimit = 0;
		upperlimit = (int)file.length();
	}

	/**
	 * Constructor, given the filename, generates a new read-only
	 * <code>FilePointer</code> bound to the values of <code>lowerlimit</code>
	 * and <code>upperlimit</code>.
	 * 
	 * @param filename	 Name of the file to create a pointer to.
	 * @param lowerlimit The lower limit within the file for which the file can
	 * 					 move to.
	 * @param upperlimit The upper limit within the file for which the file can
	 * 					 move to.
	 */
	public FilePointer(String filename, int lowerlimit, int upperlimit) {

		file = new RandomAccessFile(filename, "r");
		file.seek(lowerlimit);
		this.lowerlimit = lowerlimit;
		this.upperlimit = upperlimit;
	}

	/**
	 * Closes this <code>FilePointer</code> and any system resources associated
	 * with this file.
	 */
	public void close() {

		file.close();
	}

	/**
	 * Returns the <code>FileChannel</code> associated with this pointer.
	 * 
	 * @return This pointer's <code>FileChannel</code>.
	 */
	public FileChannel getChannel() {

		return file.getChannel();
	}

	/**
	 * Returns the <code>FileDescriptor</code> for which this file pointer is
	 * referencing.
	 * 
	 * @return File handle, from which an input/output stream can be made.
	 */
	public FileDescriptor getFileDescriptor() {

		return file.getFD();
	}

	/**
	 * Returns the offset values bound to the lower limit of this pointer's
	 * movement.
	 * 
	 * @return The lower limit imposed on this pointer.
	 */
	int offset() {

		return lowerlimit;
	}

	/**
	 * Returns the current position of the file pointer (relative to the lower
	 * limit).
	 * 
	 * @return The position of the file pointer.
	 */
	public int position() {

		return (int)file.getFilePointer() - lowerlimit;
	}

	/**
	 * Returns a single byte at the current position of the file pointer.  The
	 * pointer is then incremented by 1 byte.
	 * 
	 * @return 1 byte.
	 */
	public byte readByte() {

		return (byte)file.read();
	}

	/**
	 * Returns a single byte at the given position in the file.  The original
	 * position of the pointer is then restored afterwards.
	 * 
	 * @param position The position in the file (relative to the lower limit) to
	 * 				   read the byte from.
	 * @return 1 byte.
	 */
	public byte readByte(int position) {

		int oldpos = (int)file.getFilePointer();
		file.seek(lowerlimit + position);
		byte b = (byte)file.read();
		file.seek(oldpos);

		return b;
	}

	/**
	 * Returns a <code>ByteBuffer</code> encompassing the specified number of
	 * bytes from the current position of the <code>FilePointer</code>.  The
	 * position is then moved ahead <code>length</code> bytes.
	 * 
	 * @param length The number of bytes to read from the file.
	 * @return <code>ByteBuffer</code>, in native byte order, of the bytes read.
	 */
	public ByteBuffer readBytes(int length) {

		byte[] dest = new byte[length];
		file.read(dest);
		return BufferUtility.wrapByteBuffer(dest);
	}

	/**
	 * Returns the next int-sized value at the current position of the file
	 * pointer.  The pointer is then incremented by 4 bytes.
	 * 
	 * @return 1 int.
	 */
	public int readInt() {

		int b1 = file.read();
		int b2 = file.read();
		int b3 = file.read();
		int b4 = file.read();

		return b1 | (b2 << 8) | (b3 << 16) | (b4 << 24);
	}

	/**
	 * Returns the next line of text from the file.  Note that the way this
	 * method works is that it reads 1 byte, then appends another byte of 0 to
	 * generate a Java 16-bit character.  Luckily for me, the ASCII 8-bit
	 * characters used map directly to their 16-bit counterparts.
	 * 
	 * @return 1 line of text.
	 */
	public String readLine() {

		return file.readLine();
	}

	/**
	 * Returns the next short-sized value at the current position of the file
	 * pointer.  The pointer is then incremented by 2 bytes.
	 * 
	 * @return 1 short.
	 */
	public short readShort() {

		int b1 = file.read();
		int b2 = file.read();

		return (short)(b1 | (b2 << 8));
	}

	/**
	 * Returns the number of remaining bytes.  Essentially the difference
	 * between the current position and the upper limit.
	 * 
	 * @return The number of bytes between the current position and the upper
	 * 		   limit.
	 */
	public int remaining() {

		return upperlimit - (int)file.getFilePointer();
	}

	/**
	 * Resets the position of the <code>FilePointer</code>, back to the lower
	 * bound of the file.
	 */
	public void rewind() {

		file.seek(lowerlimit);
	}

	/**
	 * Moves the position of the file pointer to the specified position.  This
	 * position is relative to the lower bound.
	 * 
	 * @param position The new position to move the pointer to.
	 */
	public void seek(int position) {

		file.seek(lowerlimit + position);
	}

	/**
	 * Returns the size of the file this class is referencing.  Effectively, the
	 * difference between the upper and lower bounds.
	 * 
	 * @return The size of the file being manipulated.
	 */
	public int size() {

		return upperlimit - lowerlimit;
	}

	/**
	 * Writes the given <code>byte</code>-sized value at the current position of
	 * the file pointer.  The pointer is then incremented by 1 byte.
	 * 
	 * @param byteval 8-bit value to write to file at the current position of
	 * 				  the file pointer.
	 */
	public void writeByte(byte byteval) {

		file.write(byteval & 0xff);
	}

	/**
	 * Writes a series of bytes from the given <code>ByteBuffer</code> to the
	 * current position of the file pointer.  The pointer is then incremented by
	 * the number of bytes written, which is the limit of the byte buffer.
	 * 
	 * @param bytes Buffer of bytes to write to file at the current position of
	 * 				the file pointer.
	 */
	public void writeBytes(ByteBuffer bytes) {

		file.write(bytes.array(), 0, bytes.limit());
	}

	/**
	 * Writes the given <code>int</code>-sized value at the current position of
	 * the file pointer.  The pointer is then incremented by 4 bytes.
	 * 
	 * @param intval 32-bit value to write to file at the current position of
	 * 				 the file pointer.
	 */
	public void writeInt(int intval) {

		byte[] bytes = {
			(byte)(intval),
			(byte)(intval >>> 8),
			(byte)(intval >>> 16),
			(byte)(intval >>> 24)
		};

		file.write(bytes);
	}

	/**
	 * Writes the given <code>short</code>-sized value at the current position
	 * of the file pointer.  The pointer is then incremented by 2 bytes.
	 * 
	 * @param shortval 16-bit value to write to file at the current position of
	 * 				   the file pointer.
	 */
	public void writeShort(short shortval) {

		byte[] bytes = {
			(byte)(shortval),
			(byte)(shortval >>> 8)
		};

		file.write(bytes);
	}
}
