
// ===========================================
// Scanner's Java - File-related utility class
// ===========================================

package redhorizon.utilities;

import redhorizon.filetypes.ArchiveFile;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.nio.CharBuffer;
import java.nio.channels.FileChannel;

/**
 * This class deals with several file-handling tasks of Red Horizon, such as
 * copying and reading.<br>
 * <br>
 * The static methods of this class are for one-off functions like copying a
 * file, or writing some text/bytes to a file.  For functions with require
 * persistent use of a file, instead use the {@link FilePointer} utility class.
 * 
 * @author Emanuel Rabina
 */
public class FileUtility {

	/**
	 * Hidden default contructor, as this class is only ever meant to be used
	 * statically.
	 */
	private FileUtility() {
	}

	/**
	 * One-off operation, copies a file from one location to the other.
	 * 
	 * @param origpath Path to the original file.
	 * @param copypath Path of where to make a copy of the original.
	 */
	public static void copyFile(String origpath, String copypath) {

		FileChannel input  = new FileInputStream(origpath).getChannel();
		FileChannel output = new FileOutputStream(copypath).getChannel();

		input.transferTo(0, input.size(), output);

		input.close();
		output.close();
	}

	/**
	 * One-off operation, copies a file from the inside of a {@link
	 * redhorizon.filetypes.ArchiveFile} type, into the location specified.
	 * 
	 * @param origname	Name of the original file inside the container.
	 * @param container Name of the file holding the original file we're after.
	 * @param copypath	Path of where to make a copy of the original.
	 */
	public static void copyFileFromArchive(String origname, ArchiveFile container, String copypath) {

		FilePointer item = container.getItem(origname);
		FileChannel input  = item.getChannel();
		FileChannel output = new FileOutputStream(copypath).getChannel();

		input.transferTo(item.offset(), item.size(), output);

		input.close();
		output.close();
	}

	/**
	 * Writes a plain-text file to disk to the location specified, using the
	 * given character sequence.
	 * 
	 * @param path  Location of the file to write.
	 * @param chars A <code>CharBuffer</code> of characters to write to the
	 * 				file.
	 */
	public static void writeTextFile(String path, CharBuffer chars) {

		PrintWriter file = new PrintWriter(new FileWriter(path));
		file.print(chars.toString());
		file.flush();
		file.close();
	}
}
