
// ================================
// Scanner's Java - Image utilities
// ================================

package redhorizon.utilities;

import redhorizon.filetypes.ImageFile;
import redhorizon.filetypes.ImagesFile;
import redhorizon.filetypes.PaletteFile;
import redhorizon.filetypes.Paletted;
import redhorizon.filetypes.pal.PalColour;
import redhorizon.strings.ProgramStrings;
import redhorizon.strings.Common_Filetypes;

import java.nio.ByteBuffer;

/**
 * Utility class containing methods to aid with the transfer of image data from
 * a file to an OpenGL-useable texture.
 * 
 * @author Emanuel Rabina
 */
public class ImageUtility {

	/**
	 * Applies a palette to a file, using the given palette.  Used for filetypes
	 * with internal palettes.
	 * 
	 * @param source  Original colour index frame/image.
	 * @param dest	  Buffer to store the colourized data.
	 * @param palette <code>PalColour</code>[] to use.  Uses the class'
	 * 				  {@link PalColour#toArray()} method.
	 */
	public static void applyPalette(ByteBuffer source, ByteBuffer dest, PalColour[] palette) {

		for (int i = 0; i < source.limit(); i++) {
			dest.put(palette[source.get() & 0xff].toArray());
		}
		dest.rewind();
	}

	/**
	 * Applies an external palette to single paletted image.
	 * 
	 * @param source  Original colour index frame/image.
	 * @param dest	  Buffer to store the colourized data.
	 * @param format  Number of colour components per pixel.  3 for RGB, 4 for
	 * 				  RGBA.
	 * @param palette Palette file to use.
	 */
	public static void applyPalette(ByteBuffer source, ByteBuffer dest, int format, PaletteFile palette) {

		// Apply RGB/A colour values to the resulting image
		while (source.hasRemaining()) {
			dest.put(format == 3 ?
					palette.getRGB(source.get() & 0xff):
					palette.getRGBA(source.get() & 0xff));
		}
		source.rewind();
		dest.rewind();
	}

	/**
	 * Applies a palette to an array of paletted image data.
	 * 
	 * @param source  Original colour index frames/images.
	 * @param dest	  Buffers to store the colourized data.
	 * @param format  Number of colour components per pixel.  3 for RGB, 4 for
	 * 				  RGBA.
	 * @param palette Palette file to use.
	 */
	public static void applyPalette(ByteBuffer[] source, ByteBuffer[] dest, int format, PaletteFile palette) {

		for (int i = 0; i < source.length; i++) {
			applyPalette(source[i], dest[i], format, palette);
		}
	}

	/**
	 * Creates a single overall image buffer from a series of smaller image
	 * buffers.  If the image is of the paletted sort, a palette must be
	 * supplied, otherwise the palette can be <code>null</code>.  The width and
	 * height of each inner image must be the same, otherwise an exception will
	 * be thrown.
	 * 
	 * @param combinewidth	The number of images to combine horizontally.
	 * @param combineheight The number of images to combine vertically.
	 * @param imagesfile	File containing a series of images.
	 * @param palette		Palette to use for the merging (if paletted).
	 * @return Single combined image data.
	 * @throws IllegalArgumentException If the widths and heights of the source
	 * 		   images are not the same.
	 */
	public static ByteBuffer combineImages(int combinewidth, int combineheight,
			ImagesFile imagesfile, PaletteFile palette) {

		// Check image widths and heights match
		int width  = imagesfile.width(0);
		int height = imagesfile.height(0);
		for (int i = 1; i < imagesfile.numImages(); i++) {
			if (width != imagesfile.width(i) || height != imagesfile.height(i)) {
				throw new IllegalArgumentException(ProgramStrings.getText(
						Common_Filetypes.DIMENSIONS_MISMATCH_COMBINE, imagesfile.toString()));
			}
		}

		int format = imagesfile.format();
		ByteBuffer[] images = imagesfile instanceof Paletted ?
				((Paletted)imagesfile).applyPalette(palette) : imagesfile.getImages();

		// Create the full storage buffer
		ByteBuffer compilation = BufferUtility.newByteBuffer(width * height * format * imagesfile.numImages());
		for (int y = 0, rowsize = format * width; y < combineheight; y++) {

			// Append data across
			for (int i = 0; i < height; i++) {
				for (int x = 0; x < combinewidth; x++) {
					compilation.put(images[y * combineheight + x].array(), rowsize * i, rowsize);
				}
			}
		}
		compilation.rewind();
		return compilation;
	}

	/**
	 * Creates a series of images built from a single image, split using the
	 * given dimensions for the resulting images.  The source can be either
	 * indexed or non-indexed data; the result will be of the same type.<br>
	 * <br>
	 * Only works if the dimensions to split the image by cut the source image
	 * exactly, without remaining edges.  An exception is thrown if this is not
	 * the case.
	 * 
	 * @param splitwidth  Width of resulting images.
	 * @param splitheight Height of resulting images.
	 * @param numimages	  Number of images to return.  Can be used to ensure
	 * 					  blank images aren't in the result.
	 * @param imagefile	  File to source images from.
	 * @throws IllegalArgumentException If the source image cannot be split
	 * 		   evenly as requested.
	 */
	public static ByteBuffer[] splitImage(int splitwidth, int splitheight, int numimages, ImageFile imagefile) {

		return splitImage(splitwidth, splitheight, numimages,
				imagefile.width(), imagefile.height(), imagefile.format(), imagefile.getImage());
	}

	/**
	 * Creates a series of images built from a single image, split using the
	 * given dimensions for the resulting images.  The source can be either
	 * indexed or non-indexed data; the result will be of the same type.<br>
	 * <br>
	 * Only works if the dimensions to split the image by cut the source image
	 * exactly, without remaining edges.  An exception is thrown if this is not
	 * the case.
	 * 
	 * @param splitwidth  Width of resulting images.
	 * @param splitheight Height of resulting images.
	 * @param numimages	  Number of images to return.  Can be used to ensure
	 * 					  blank images aren't in the result.
	 * @param srcwidth	  Width of the source image.
	 * @param srcheight	  Height of the source image.
	 * @param format	  Number of colour components in the source image.
	 * @param srcimage	  File to source images from.
	 * @throws IllegalArgumentException If the source image cannot be split
	 * 		   evenly as requested.
	 */
	public static ByteBuffer[] splitImage(int splitwidth, int splitheight, int numimages,
			int srcwidth, int srcheight, int format, ByteBuffer srcimage) {

		// Check result dimensions leave no remainders
		if ((srcwidth % splitwidth != 0) || (srcheight % splitheight != 0)) {
			throw new IllegalArgumentException(ProgramStrings.getText(
					Common_Filetypes.DIMENSIONS_MISMATCH_SPLIT));
		}

		// Create split images
		ByteBuffer[] compilation = new ByteBuffer[numimages];
		for (int i = 0, row = 0, col = 0, numimghor = srcwidth / splitwidth; i < numimages; ) {
			ByteBuffer splitimage = BufferUtility.newByteBuffer(splitwidth * splitheight * format);

			for (int y = 0; y < splitheight; y++) {
				srcimage.position((row + col) + (y * srcwidth));
				srcimage.get(splitimage.array(), y * splitwidth, splitwidth);
			}
			splitimage.rewind();
			compilation[i] = splitimage;

			// Adjust current data co-ordinates in source image
			i++;
			if (i % numimghor == 0) {
				row += srcwidth * splitheight * format;
			}
			col = (i % numimghor) * splitwidth;
		}
		srcimage.rewind();

		return compilation;
	}
}
