
// ==========================================
// Scanner's Java - GUI-builder utility class
// ==========================================

package redhorizon.utilities;


import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;

import redhorizon.misc.geometry.Rectangle4i;

/**
 * Essentially a set of common routines used for constructing the OS-based
 * GUI/windows used in the program.  Mainly layout methods.
 * 
 * @author Emanuel Rabina
 */
public class SWTLayouts {

	/**
	 * Used to specify horizontal alignment over the <code>GridData</code>
	 * component.
	 */
	public static enum HAlign {

		LEFT   (SWT.LEFT),
		CENTER (SWT.CENTER),
		RIGHT  (SWT.RIGHT),
		FILL   (SWT.FILL),
		NONE   (SWT.NONE);

		public final int align;

		/**
		 * Constructor, assigns the alignment bit.
		 * 
		 * @param align SWT equivalent of the alignment enum.
		 */
		private HAlign(int align) {

			this.align = align;
		}
	}

	/**
	 * Used to specify vertical alignment over the <code>GridData</code>
	 * component.
	 */
	public static enum VAlign {

		TOP    (SWT.TOP),
		MIDDLE (SWT.CENTER),
		BOTTOM (SWT.BOTTOM),
		FILL   (SWT.FILL),
		NONE   (SWT.NONE);

		public final int align;

		/**
		 * Constructor, assigns the alignment bit.
		 * 
		 * @param align SWT equivalent of the alignment enum.
		 */
		private VAlign(int align) {

			this.align = align;
		}
	}

	/**
	 * Hidden default contructor, as this class is only ever meant to be used
	 * statically.
	 */
	private SWTLayouts() {
	}

	/**
	 * Creates a <code>GridData</code> layout component with the given width and
	 * height.
	 * 
	 * @param size Dimensions of the component.
	 * @return A <code>GridData</code> with the specified parameters.
	 */
	public static GridData createLayoutData(Rectangle4i size) {

		return createLayoutData(size, HAlign.NONE, VAlign.NONE, false, false, 1, 1);
	}

	/**
	 * Creates a <code>GridData</code> layout component with the given
	 * horizontal and vertical alignments.
	 * 
	 * @param halign
	 * @param valign
	 * @return A <code>GridData</code> with the specified parameters.
	 */
	public static GridData createLayoutData(HAlign halign, VAlign valign) {

		return createLayoutData(new Rectangle4i(SWT.DEFAULT, SWT.DEFAULT),
				halign, valign, false, false, 1, 1);
	}

	/**
	 * Creates a <code>GridData</code> layout component with the given
	 * horizontal and vertical fills.
	 * 
	 * @param hfill
	 * @param vfill
	 * @return A <code>GridData</code> with the specified parameters.
	 */
	public static GridData createLayoutData(boolean hfill, boolean vfill) {

		return createLayoutData(new Rectangle4i(SWT.DEFAULT, SWT.DEFAULT),
				HAlign.NONE, VAlign.NONE, hfill, vfill, 1, 1);
	}

	/**
	 * Creates a <code>GridData</code> layout component with the given
	 * horizontal and vertical spans.
	 * 
	 * @param hspan
	 * @param vspan
	 * @return A <code>GridData</code> with the specified parameters.
	 */
	public static GridData createLayoutData(int hspan, int vspan) {

		return createLayoutData(new Rectangle4i(SWT.DEFAULT, SWT.DEFAULT),
				HAlign.NONE, VAlign.NONE, false, false, hspan, vspan);
	}

	/**
	 * Creates a <code>GridData</code> layout component with the given width,
	 * height, and horizontal and vertical alignments.
	 * 
	 * @param size	 Dimensions of the component.
	 * @param halign Horizontal alignment of the component.
	 * @param valign Vertical alignment of the component.
	 * @return A <code>GridData</code> with the specified parameters.
	 */
	public static GridData createLayoutData(Rectangle4i size, HAlign halign, VAlign valign) {

		return createLayoutData(size, halign, valign, false, false, 1, 1);
	}

	/**
	 * Creates a <code>GridData</code> layout component with the given width,
	 * height, horizontal and vertical fills.
	 * 
	 * @param size Dimensions of the component.
	 * @param hfill
	 * @param vfill
	 * @return A <code>GridData</code> with the specified parameters.
	 */
	public static GridData createLayoutData(Rectangle4i size, boolean hfill, boolean vfill) {

		return createLayoutData(size, HAlign.NONE, VAlign.NONE, hfill, vfill, 1, 1);
	}

	/**
	 * Creates a <code>GridData</code> layout component with the given width,
	 * height, horizontal and vertical spans.
	 * 
	 * @param size Dimensions of the component.
	 * @param hspan
	 * @param vspan
	 * @return A <code>GridData</code> with the specified parameters.
	 */
	public static GridData createLayoutData(Rectangle4i size, int hspan, int vspan) {

		return createLayoutData(size, HAlign.NONE, VAlign.NONE, false, false, hspan, vspan);
	}

	/**
	 * Creates a <code>GridData</code> layout component with the given
	 * horizontal and vertical alignments and fills.
	 * 
	 * @param halign Horizontal alignment of the component.
	 * @param valign Vertical alignment of the component.
	 * @param hfill
	 * @param vfill
	 * @return A <code>GridData</code> with the specified parameters.
	 */
	public static GridData createLayoutData(HAlign halign, VAlign valign, boolean hfill, boolean vfill) {

		return createLayoutData(new Rectangle4i(SWT.DEFAULT, SWT.DEFAULT),
				halign, valign, hfill, vfill, 1, 1);
	}

	/**
	 * Creates a <code>GridData</code> layout component with the given
	 * horizontal and vertical spans and alignments.
	 * 
	 * @param halign Horizontal alignment of the component.
	 * @param valign Vertical alignment of the component.
	 * @param hspan
	 * @param vspan
	 * @return A <code>GridData</code> with the specified parameters.
	 */
	public static GridData createLayoutData(HAlign halign, VAlign valign, int hspan, int vspan) {

		return createLayoutData(new Rectangle4i(SWT.DEFAULT, SWT.DEFAULT),
				halign, valign, false, false, hspan, vspan);
	}

	/**
	 * Creates a <code>GridData</code> layout component with the given
	 * horizontal and vertical fills and spans.
	 * 
	 * @param hfill
	 * @param vfill
	 * @param hspan
	 * @param vspan
	 * @return A <code>GridData</code> with the specified parameters.
	 */
	public static GridData createLayoutData(boolean hfill, boolean vfill, int hspan, int vspan) {

		return createLayoutData(new Rectangle4i(SWT.DEFAULT, SWT.DEFAULT),
				HAlign.NONE, VAlign.NONE, hfill, vfill, hspan, vspan);
	}

	/**
	 * Creates a <code>GridData</code> layout component with the given width,
	 * height, alignments, and fills.
	 * 
	 * @param size	  Dimensions of the component.
	 * @param halign  Horizontal alignment of the component.
	 * @param valign  Vertical alignment of the component.
	 * @param hfill
	 * @param vfill
	 * @return A <code>GridData</code> with the specified parameters.
	 */
	public static GridData createLayoutData(Rectangle4i size, HAlign halign, VAlign valign,
			boolean hfill, boolean vfill) {

		return createLayoutData(size, halign, valign, hfill, vfill, 1, 1);
	}

	/**
	 * Creates a <code>GridData</code> layout component with the given width,
	 * height, alignments, and fills.
	 * 
	 * @param size	 Dimensions of the component.
	 * @param halign Horizontal alignment of the component.
	 * @param valign Vertical alignment of the component.
	 * @param hspan
	 * @param vspan
	 * @return A <code>GridData</code> with the specified parameters.
	 */
	public static GridData createLayoutData(Rectangle4i size, HAlign halign, VAlign valign,
			int hspan, int vspan) {

		return createLayoutData(size, halign, valign, false, false, hspan, vspan);
	}

	/**
	 * Creates a <code>GridData</code> layout component with the given width,
	 * height, alignments, and fills.
	 * 
	 * @param size	Dimensions of the component.
	 * @param hfill
	 * @param vfill
	 * @param hspan
	 * @param vspan
	 * @return A <code>GridData</code> with the specified parameters.
	 */
	public static GridData createLayoutData(Rectangle4i size,
			boolean hfill, boolean vfill, int hspan, int vspan) {

		return createLayoutData(size, HAlign.NONE, VAlign.NONE, hfill, vfill, hspan, vspan);
	}

	/**
	 * Creates a <code>GridData</code> layout component with the given width,
	 * height, alignments, and fills.
	 * 
	 * @param halign  Horizontal alignment of the component.
	 * @param valign  Vertical alignment of the component.
	 * @param hfill
	 * @param vfill
	 * @param hspan
	 * @param vspan
	 * @return A <code>GridData</code> with the specified parameters.
	 */
	public static GridData createLayoutData(HAlign halign, VAlign valign,
			boolean hfill, boolean vfill, int hspan, int vspan) {

		return createLayoutData(new Rectangle4i(SWT.DEFAULT, SWT.DEFAULT),
				halign, valign, hfill, vfill, hspan, vspan);
	}

	/**
	 * Creates a <code>GridData</code> layout component with the given width,
	 * height, alignment, horizontal and vertical fills and spans.
	 * 
	 * @param size	  Dimensions of the component.
	 * @param halign  Horizontal alignment of the component.
	 * @param valign  Vertical alignment of the component.
	 * @param hfill
	 * @param vfill
	 * @param hspan
	 * @param vspan
	 * @return A <code>GridData</code> with the specified parameters.
	 */
	public static GridData createLayoutData(Rectangle4i size, HAlign halign, VAlign valign,
			boolean hfill, boolean vfill, int hspan, int vspan) {

		GridData griddata = new GridData(halign.align, valign.align, hfill, vfill, hspan, vspan);
		griddata.widthHint  = size.getWidth();
		griddata.heightHint = size.getHeight();

		return griddata;
	}

	/**
	 * Generic layout, all parameters of the layout can be specified.
	 * 
	 * @param numcolumns Number of columns in this layout.
	 * @param equalspace Whether or not the columns should be of the same width.
	 * @param margins	 Rectangle defining margin sizes on all sides.
	 * @param spacing	 Rectangle defining gaps between widgets, uses the width
	 * 					 and height attributes of the rectangle.
	 * @return A <code>GridLayout</code> with the passed settings.
	 */
	public static GridLayout createLayout(int numcolumns, boolean equalspace,
			Rectangle4i margins, Rectangle4i spacing) {

		GridLayout gridlayout = new GridLayout(numcolumns, equalspace);
		gridlayout.marginWidth  = 0;
		gridlayout.marginHeight = 0;
		gridlayout.marginTop    = margins.getY();
		gridlayout.marginLeft   = margins.getX();
		gridlayout.marginRight  = margins.getWidth();
		gridlayout.marginBottom = margins.getHeight();
		gridlayout.horizontalSpacing = spacing.getWidth();
		gridlayout.verticalSpacing   = spacing.getHeight();

		return gridlayout;
	}

	/**
	 * Creates a new group/composite-level layout manager based on the
	 * <code>GridLayout</code>.  The defaults for this layout are:<br>
	 * <li>Columns: 1</li>
	 * <li>Equal Width: yes</li>
	 * <li>Margin Width: 7</li>
	 * <li>Margin Height: 7</li>
	 * <li>Vertical Spacing: 7</li>
	 * <li>Horizontal Spacing: 7</li>
	 * 
	 * @return A new <code>GridLayout</code> based on the above defaults.
	 */
	public static GridLayout createLayoutInternal() {

		return createLayoutInternal(1, true);
	}

	/**
	 * Creates a new internal layout manager based on the
	 * <code>GridLayout</code>.  The defaults for this layout are:<br>
	 * <li>Margin Width: 7</li>
	 * <li>Margin Height: 7</li>
	 * <li>Vertical Spacing: 7</li>
	 * <li>Horizontal Spacing: 7</li>
	 * 
	 * @param numcolumns Number of columns in this layout.
	 * @param equalspace Whether or not the columns should be of the same width.
	 * @return A new <code>GridLayout</code> based on the above settings.
	 */
	public static GridLayout createLayoutInternal(int numcolumns, boolean equalspace) {

		return createLayout(numcolumns, equalspace, new Rectangle4i(5,5,5,5), new Rectangle4i(5,5));
	}

	/**
	 * Creates a new group/composite-level layout manager based on the
	 * <code>GridLayout</code>.  The defaults for this layout are:<br>
	 * <li>Columns: 1</li>
	 * <li>Equal Width: yes</li>
	 * <li>Margin Width: 5</li>
	 * <li>Margin Height: 5</li>
	 * <li>Vertical Spacing: 5</li>
	 * <li>Horizontal Spacing: 5</li>
	 * 
	 * @return A new <code>GridLayout</code> based on the above defaults.
	 */
	public static GridLayout createLayoutGroup() {

		return createLayoutGroup(1, true);
	}

	/**
	 * Creates a new group/composite-level layout manager based on the
	 * <code>GridLayout</code>.  The defaults for this layout are:<br>
	 * <li>Margin Width: 5</li>
	 * <li>Margin Height: 5</li>
	 * <li>Vertical Spacing: 5</li>
	 * <li>Horizontal Spacing: 5</li>
	 * 
	 * @param numcolumns Number of columns in this layout.
	 * @param equalspace Whether or not the columns should be of the same width.
	 * @return A new <code>GridLayout</code> based on the above settings.
	 */
	public static GridLayout createLayoutGroup(int numcolumns, boolean equalspace) {

		return createLayout(numcolumns, equalspace, new Rectangle4i(7,7,7,7), new Rectangle4i(7,7));
	}

	/**
	 * Creates a new shell-level layout manager based on the
	 * <code>GridLayout</code>.  The defaults for this layout are:<br>
	 * <li>Columns: 1</li>
	 * <li>Equal Width: yes</li>
	 * <li>Margin Width: 10</li>
	 * <li>Margin Height: 10</li>
	 * <li>Vertical Spacing: 7</li>
	 * <li>Horizontal Spacing: 7</li>
	 * 
	 * @return A new <code>GridLayout</code> based on the above defaults.
	 */
	public static GridLayout createLayoutShell() {

		return createLayout(1, true, new Rectangle4i(0,0,0,0), new Rectangle4i(0,0));
	}
}
