
// =======================================================
// Scanner's AspectJ - Ensures the Converter runs smoothly
// =======================================================

package redhorizon.utilities.converter;

import redhorizon.filetypes.FileImplementation;
import redhorizon.strings.ProgramStrings;
import redhorizon.strings.Utilities_Converter;

import org.aspectj.lang.SoftException;
import java.io.FileNotFoundException;
import java.lang.reflect.InvocationTargetException;

/**
 * Performs tasks and checks on the {@link Converter}, including:<br>
 *  - ensuring the right parameters are used<br>
 *  - disallowing invalid file conversions<br>
 *  - softening checked exceptions
 * 
 * @author Emanuel Rabina
 */
public aspect AspectConverter {

	/**
	 * Get to the root cause of program exceptions and fire-off an error message
	 * appropriate for the error.
	 * 
	 * @param args Array of command-line arguments.
	 */
	void around(String[] args):
		execution(public static void Converter.main(String[])) && args(args) {

		try {
			// Check for proper number of parameters
			if (Integer.parseInt(args[args.length - 1]) != args.length - 1) {
				throw new IllegalArgumentException();
			}

			// Proceed with conversion
			proceed(args);
		}
		catch (IllegalArgumentException iae) {
			System.exit(1);
		}
		catch (SoftException se) {
			Throwable cause = getRootCause(se);

			if (cause instanceof FileNotFoundException) {
				System.out.println(ProgramStrings.getText(Utilities_Converter.FILE_NOT_FOUND));
			}
			else if (cause instanceof NoSuchMethodException) {
				System.out.println(ProgramStrings.getText(Utilities_Converter.UNSUPPORTED_CONVERSION));
			}
			else {
				cause.printStackTrace();
				System.out.println(cause.getMessage());
			}
		}
	}

	/**
	 * Recursively finds the root cause of an exception.
	 * 
	 * @param exception The exception that was caught.
	 * @return The root cause of the exception.
	 */
	private static Throwable getRootCause(Throwable exception) {

		Throwable cause = exception.getCause();
		if (cause != null) {
			return getRootCause(cause);
		}
		return exception;
	}

	/**
	 * Softening of exceptions over the building of the FROM/TO filetypes.
	 */
	pointcut buildFiles():
		execution(private static FileImplementation Converter.build*(..));

	declare soft: ClassNotFoundException:    buildFiles();
	declare soft: IllegalAccessException:    buildFiles();
	declare soft: InstantiationException:    buildFiles();
	declare soft: InvocationTargetException: buildFiles();
	declare soft: NoSuchMethodException:     buildFiles();
}
