
// ===========================================
// Scanner's Java - Red Horizon file converter
// ===========================================

package redhorizon.utilities.converter;

import redhorizon.filetypes.FileImplementation;
import redhorizon.settings.Settings;
import redhorizon.strings.ProgramStrings;
import redhorizon.strings.Utilities_Converter;

import java.lang.reflect.Constructor;

/**
 * File converter utility, used to build files of one type from files of another
 * type.  A part of the Red Horizon utilities project and relies on the
 * {@link redhorizon.filetypes} package.<br>
 * <br>
 * The <code>Converter</code> can only save to files of the filetypes package
 * which have implemented the {@link FileImplementation#saveToFile()} method.
 * On top of that, only some file conversions are allowed; any attempt to save
 * or perform an unsupported conversions will result in a
 * {@link ConverterException}.<br>
 * <br>
 * Some credit goes to C&C community member Nyerguds, who initially requested a
 * Dune 2 SHP file writer.  That was the starting-point for the Red Horizon
 * utilities project.
 * 
 * @author Emanuel Rabina
 */
public class Converter {

	/**
	 * Hidden default constructor, as this class is only ever meant to be used
	 * statically.
	 */
	private Converter() {
	}

	/**
	 * Entry point for the Red Horizon file converter utility. Parameters are
	 * expected as follows:<br>
	 *  - short name of the FROM type<br>
	 *  - short name of the TO type<br>
	 *  - name of the file to get the data FROM<br>
	 *  - name of the file to save the data TO<br>
	 * <br>
	 * eg: <code>java Converter pcx.PcxFile shp.ShpFileDune2 pcxfile.pcx shpfile.shp</code><br>
	 * <br>
	 * Anything afterwards is an optional parameter for the TO file and works on
	 * a per-conversion basis.
	 * 
	 * @param args Command-line parameters.
	 */
	public static void main(String[] args) {

		// Extract parameters
		String fromtype = args[0];
		String totype   = args[1];
		String fromfile = args[2];
		String tofile   = args[3];

		// Build file saving parameters
		String[] params;
		if (args.length > 4) {
			params = new String[args.length - 4];
			System.arraycopy(args, 4, params, 0, params.length);
		}
		else {
			params = null;
		}

		// Init settings, strings manager, add converter strings
		Settings.init();
		ProgramStrings.init();
		ProgramStrings.addMessageBundle(Utilities_Converter.class);

		// Read the requested file type
		System.out.println(ProgramStrings.getText(Utilities_Converter.LOADING_INPUT));
		FileImplementation fromfileobj = buildFromFile(fromtype, fromfile);
		fromfileobj.readFromFile();
		System.out.println(fromfileobj + "\n");

		// Save to the requested file type
		System.out.println(ProgramStrings.getText(Utilities_Converter.SAVING_OUTPUT));
		FileImplementation tofileobj = buildToFile(totype, tofile, fromfileobj, params);
		tofileobj.saveToFile();
		System.out.println(tofileobj + "\n");

		// Complete
		System.out.println(ProgramStrings.getText(Utilities_Converter.CONVERSION_COMPLETE));
	}

	/**
	 * Builds the implementation for the FROM file.
	 * 
	 * @param fromtype Short name of the filetype to instantiate.
	 * @param fromfile Name of the file this type will be built from.
	 * @return Object of the given filetype.
	 */
	private static FileImplementation buildFromFile(String fromtype, String fromfile) {

		Class<?> fromfileimpl = Class.forName("redhorizon.filetypes." + fromtype);
		Constructor<?> fromfileconst = fromfileimpl.getConstructor(String.class);
		return (FileImplementation)fromfileconst.newInstance(fromfile);
	}

	/**
	 * Builds the implementation for the TO file.
	 * 
	 * @param totype  Short name of the filetype to instantiate.
	 * @param tofile  Name of the file this type will be built to.
	 * @param fromobj Object of the FROM file.
	 * @param params  Additional parameters to use for saving the TO file.
	 * 				  <code>null</code> if parameters aren't used.
	 * @return Object of the given filetype.
	 */
	private static FileImplementation buildToFile(String totype, String tofile,
			FileImplementation fromobj, String... params) {

		Class<?> tofileclass = Class.forName("redhorizon.filetypes." + totype);
		Constructor<?> tofileconst;
		FileImplementation tofileimpl;

		if (params == null) {
			tofileconst = tofileclass.getConstructor(String.class, fromobj.getClass());
			tofileimpl = (FileImplementation)tofileconst.newInstance(tofile, fromobj);
		}
		else {
			tofileconst = tofileclass.getConstructor(String.class, fromobj.getClass(), params.getClass());
			tofileimpl = (FileImplementation)tofileconst.newInstance(tofile, fromobj, params);
		}

		return tofileimpl;
	}
}
